import os
import time
from logging import raiseExceptions

import numpy as np
import pandas as pd
import scipy.sparse as ssp
import scipy.stats as stats
import statsmodels.sandbox.stats.multicomp
from ete3 import Tree
from matplotlib import pyplot as plt
from scipy.cluster import hierarchy

# from plotnine import *
from sklearn.manifold import SpectralEmbedding

from cospar.tool import _clone, _utils

from .. import help_functions as hf
from .. import logging as logg
from .. import settings
from ..help_functions import _docs
from ..help_functions._docs import _doc_params


def fate_hierarchy(
    adata,
    selected_fates: list = None,
    source: str = "X_clone",
    selected_times: list = None,
    method: str = "SW",
    normalize: bool=True,
    ignore_cell_number: bool = False,
):
    """
    Build fate hierarchy or lineage trees

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    selected_fates: `list`, optional (default: all fates)
        List of cluster ids consistent with adata.obs['state_info'].
        It allows a nested list, where we merge clusters within
        each sub-list into a mega-fate cluster.
    source: `str`, optional (default: 'transition_map')
        Choices: {'X_clone', 'transition_map',
        'intraclone_transition_map',...}. If set to be 'clone', use only the clonal
        information. If set to be any of the precomputed transition map, use the
        transition map to compute the fate coupling. The actual available
        map depends on adata itself, which can be accessed at adata.uns['available_map']
    selected_times: `list`, optional (default: all)
        A list of time points to further restrict the cell states to plot.
        The default choice is not to constrain the cell states to show.
    method: `str`, optional (default: 'SW')
        Method to normalize the coupling matrix: {'SW', 'Weinreb'}.
    normalize:
        We normalize the count matrix first within each cluster, then, within each clone separately for each time point. This was set to be off in the original CoSpar publication.
    ignore_cell_number:
        Ignore the cell number of a clone within a cluster. i.e., binarize a clone's
        contribution towards a cluster. This only works when 'source=X_clone'
        This biases towards large clusters, which will host many clones (most of them just
        a few cells)

    Returns
    -------
    Results stored at adata.uns[f"fate_hierarchy_{source}"]
    """

    state_info = np.array(adata.obs["state_info"])
    if len(set(state_info))==1:
        print('Only one cluster, skip hierarchy generation')
    else:
        mega_cluster_list, valid_fate_list, __, __ = hf.analyze_selected_fates(
            state_info, selected_fates
        )

        fate_N = len(valid_fate_list)
        X_history = []
        merged_pairs_history = []
        node_names_history = []
        node_groups = {i: [i] for i in range(fate_N)}

        parent_map = {}
        selected_fates_tmp = []
        for xx in valid_fate_list:
            if type(xx) is not list:
                xx = [xx]
            selected_fates_tmp.append(xx)
        node_names = list(range(fate_N))
        next_node = fate_N

        counter = 0
        while len(node_names) > 2:
            counter += 1
            fate_N_tmp = len(selected_fates_tmp)
            node_names_history.append(node_names)
            fate_coupling(
                adata,
                selected_fates=selected_fates_tmp,
                source=source,
                selected_times=selected_times,
                method=method,
                normalize=normalize,
                silence=True,
                ignore_cell_number=ignore_cell_number,
            )
            X_coupling = adata.uns[f"fate_coupling_{source}"]["X_coupling"]
            if counter == 1:
                fate_names = adata.uns[f"fate_coupling_{source}"]["fate_names"]

            X_history.append(np.array(X_coupling))
            floor = X_coupling.min() - 100
            for i in range(X_coupling.shape[0]):
                for j in range(X_coupling.shape[1]):
                    if i >= j:
                        X_coupling[i, j] = floor

            ii = np.argmax(X_coupling.max(1))
            jj = np.argmax(X_coupling.max(0))
            merged_pairs_history.append((ii, jj))
            node_groups[next_node] = (
                node_groups[node_names[ii]] + node_groups[node_names[jj]]
            )

            parent_map[node_names[ii]] = next_node
            parent_map[node_names[jj]] = next_node

            ix = np.min([ii, jj])
            node_names = [
                n for n in node_names if not n in np.array(node_names)[np.array([ii, jj])]
            ]
            new_ix = np.array([i for i in range(fate_N_tmp) if not i in [ii, jj]])

            if len(new_ix) == 0:
                break
            new_fate = selected_fates_tmp[ii] + selected_fates_tmp[jj]
            selected_fates_tmp_1 = [selected_fates_tmp[new_ix[xx]] for xx in range(ix)]
            selected_fates_tmp_1.append(new_fate)
            for xx in range(ix, fate_N_tmp - 2):
                selected_fates_tmp_1.append(selected_fates_tmp[new_ix[xx]])
            selected_fates_tmp = selected_fates_tmp_1
            node_names.insert(ix, next_node)
            next_node += 1

        for i in node_names:
            parent_map[i] = next_node

        node_mapping = {}
        for key, value in node_groups.items():
            node_mapping[key] = [fate_names[xx] for xx in value]

        history = (X_history, merged_pairs_history, node_names_history)
        t = _utils.convert_to_tree(parent_map, fate_names)

        adata.uns[f"fate_hierarchy_{source}"] = {
            "parent_map": parent_map,
            "node_mapping": node_mapping,
            "history": history,
            "fate_names": fate_names,
            "tree": t,
        }
        logg.info(f"Results saved as dictionary at adata.uns['fate_hierarchy_{source}']")


def fate_coupling(
    adata,
    selected_fates=None,
    source="transition_map",
    selected_times=None,
    fate_map_method="sum",
    method="SW",
    normalize=True,
    silence=False,
    ignore_cell_number: bool = False,
    fate_normalize_source="X_clone",
    select_clones_with_fates: list = None,
    select_clones_without_fates: list = None,
    select_clones_mode: str = "or",
    **kwargs,
):
    """
    Compute fate coupling determined by the transition map.

    We use the fate map :math:`P_i(\mathcal{C}_l)` towards a set of
    fate clusters :math:`\{\mathcal{C}_l, l=0,1,2...\}` to compute the
    fate coupling :math:`Y_{ll'}`.

    * If method='SW': we first obtain :math:`Y_{ll'}=\sum_i P_i(\mathcal{C}_l)P_i(\mathcal{C}_{l'})`.
      Then, we normalize the the coupling: :math:`Y_{ll'}\leftarrow Y_{ll'}/\sqrt{Y_{ll}Y_{l'l'}}`.

    * If method='Weinreb', we calculate the normalized
      covariance as in :func:`~cospar.tl.get_normalized_covariance`

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    selected_fates: `list`, optional (default: all fates)
        List of cluster ids consistent with adata.obs['state_info'].
        It allows a nested list, where we merge clusters within
        each sub-list into a mega-fate cluster.
    source: `str`, optional (default: 'transition_map')
        Choices: {'X_clone', 'transition_map',
        'intraclone_transition_map',...}. If set to be 'clone', use only the clonal
        information. If set to be any of the precomputed transition map, use the
        transition map to compute the fate coupling. The actual available
        map depends on adata itself, which can be accessed at adata.uns['available_map']
    selected_times: `list`, optional (default: all)
        A list of time points to further restrict the cell states to plot.
        The default choice is not to constrain the cell states to show.
    fate_map_method: `str`, optional (default: 'sum')
        Method to obtain the fate probability map :math:`P_i(\mathcal{C})` towards a set
        of states annotated with fate :math:`\mathcal{C}`. Available options:
        {'sum', 'norm-sum'}. See :func:`.fate_map`.
    method: `str`, optional (default: 'SW')
        Method to normalize the coupling matrix: {'SW', 'Weinreb','Jaccard'}.
    normalize:
        We normalize the count matrix first within each cluster, then, within each clone separately for each time point. This was set to be off in the original CoSpar publication.
    silence:
        Suppress information printing.
    ignore_cell_number:
        Ignore the cell number of a clone within a cluster. i.e., binarize a clone's
        contribution towards a cluster. This only works when 'source=X_clone'
        This biases towards large clusters, which will host many clones (most of them just
        a few cells)
    fate_normalize_source:
        Source for cluster-wise normalization: {'X_clone','state_info'}. 'X_clone': directly row-normalize coarse_X_clone; 'state_info': compute each cluster size directly, and then normalize coarse_X_clone. The latter method is useful if we have single-cell resolution for each fate.
    select_clones_with_fates: list = None,
        Select clones that labels fates from this list.
    select_clones_without_fates: list = None,
        Exclude clones that labels fates from this list.
    select_clones_mode: str = {'or','and'}
        Logic rule for selection.

    Returns
    -------
    X_coupling and the corresponding fate_names index are stored as a dictionary
    at adata.uns[f"fate_coupling_{source}"]={"X_coupling": X_coupling, 'fate_names': fate_names}
    """

    hf.check_available_map(adata)
    time_info = np.array(adata.obs["time_info"])
    choices = list(adata.uns["available_map"]) + ["X_clone","X_similarity"]
    if source not in choices:
        raise ValueError(f"source should be among {choices}")
    elif source == "X_clone":
        coarse_X_clone, fate_names = _clone.coarse_grain_clone_over_cell_clusters(
            adata,
            selected_times=selected_times,
            selected_fates=selected_fates,
            normalize=normalize,
            fate_normalize_source=fate_normalize_source,
            select_clones_with_fates=select_clones_with_fates,
            select_clones_without_fates=select_clones_without_fates,
            select_clones_mode=select_clones_mode,
            **kwargs,
        )
        exclude_idx = coarse_X_clone.sum(1) == 0
        if np.sum(exclude_idx) > 0:
            coarse_X_clone = coarse_X_clone[~exclude_idx]
            fate_names = np.array(fate_names)
            logg.warn(
                f"{fate_names[exclude_idx]} is excluded due to lack of any clones"
            )
            fate_names = fate_names[~exclude_idx]

        if ignore_cell_number:
            coarse_X_clone = (coarse_X_clone > 0).astype(int)
        X_coupling = _utils.get_normalized_covariance(coarse_X_clone.T, method=method)
    elif source == "X_similarity":
        
        time_info = np.array(adata.obs["time_info"])
        if selected_times is not None:
            if type(selected_times) is not list:
                selected_times = [selected_times]
        sp_idx_0 = hf.selecting_cells_by_time_points(time_info, selected_times)
        state_annote = adata[sp_idx_0].obs["state_info"]
        (
            mega_cluster_list,
            valid_fate_list,
            fate_array_flat,
            sel_index_list,
        ) = hf.analyze_selected_fates(state_annote, selected_fates)

        Smatrix_0=adata.obsm['X_similarity']
        Smatrix=Smatrix_0[sp_idx_0][:,sp_idx_0]
        X_coupling = np.zeros((len(mega_cluster_list), len(mega_cluster_list)))
        for j, idx_0 in enumerate(sel_index_list):
            for k, idx_1 in enumerate(sel_index_list):
                # print('S',Smatrix.shape)
                # print('idx_0',idx_0)
                # print('idx_1',idx_1)
                tmp=Smatrix[idx_0][:,idx_1]
                X_coupling[j, k]=np.median(tmp.flatten())
                # print(tmp)

                # # Extract the upper triangular part of the matrix
                # upper_triangle = tmp[np.triu_indices(tmp.shape[0], k = 1)]
                # # Calculate the average value of the upper triangle
                # X_coupling[j, k] = upper_triangle.mean()

        fate_names = mega_cluster_list

    else:
        cell_id_t1 = adata.uns["Tmap_cell_id_t1"]
        sp_idx = hf.selecting_cells_by_time_points(
            time_info[cell_id_t1], selected_times
        )

        (
            fate_map,
            mega_cluster_list,
            __,
            __,
            __,
            __,
            __,
        ) = _utils.compute_fate_probability_map(
            adata,
            selected_fates=selected_fates,
            used_Tmap=source,
            map_backward=True,
            method=fate_map_method,
        )
        fate_idx = fate_map.sum(0) > 0
        if (len(mega_cluster_list) == 0) or (np.sum(sp_idx) == 0):
            raise ValueError("No cells selected. Computation aborted!")

        else:
            X_coupling = _utils.get_normalized_covariance(
                fate_map[sp_idx][:, fate_idx], method=method
            )

        if np.sum(~fate_idx) > 0:
            logg.warn(
                f"{mega_cluster_list[~fate_idx]} are ignored due to lack of cells"
            )
        fate_names = mega_cluster_list[fate_idx]

    adata.uns[f"fate_coupling_{source}"] = {
        "X_coupling": X_coupling,
        "fate_names": fate_names,
    }

    if not silence:
        logg.info(f"Results saved as dictionary at adata.uns['fate_coupling_{source}']")


def pvalue_for_fate_coupling(
    adata_orig, selected_fates=None, max_N_simutation=1000, normalize=False
):
    """
    Compute the pvalue for the fate coupling matrix. Pvalues are stored at adata.uns['fate_coupling_X_clone']

    Returns:
        adata: a new adata that contains the pvalue and the observed X_coupling matrix
        X_coupling_random: X_coupling matrix from all permutation test
    """
    (
        mega_cluster_list,
        valid_fate_list,
        fate_array_flat,
        sel_index_list,
    ) = hf.analyze_selected_fates(adata_orig.obs["state_info"], selected_fates)
    all_sel_idx = sel_index_list.sum(0) > 0

    # only select cells with clonal barcodes and within targeted clusters, so that
    # the permutation does not affect the clone size and the number of clonal cells in each cell type
    adata = adata_orig[(adata_orig.obsm["X_clone"].A.sum(1) > 0) & all_sel_idx]
    adata.obs["time_info"] = pd.Categorical(adata.obs["time_info"].astype(str))
    ordered_index = adata.obs["time_info"].sort_values(ascending=True).index
    adata = adata[ordered_index]
    time_info_order = sorted(
        adata.obs["time_info"].unique()
    )  # this is by default in ascending order

    from tqdm import tqdm

    from cospar import logging as logg

    adata_rand = adata.copy()
    adata_rand.uns["data_des"] = ["rand"]
    X_clone = adata.obsm["X_clone"].A

    # compute the actual coupling matrix
    fate_coupling(
        adata, selected_fates=selected_fates, source="X_clone", normalize=normalize
    )

    old_verbosity = settings.verbosity
    settings.verbosity = 1
    X_coupling_true = adata.uns["fate_coupling_X_clone"]["X_coupling"]

    # compute the permuted coupling matrix. Permute separately within each time point, as the normalization is within each time point
    X_coupling_random = []
    for _ in tqdm(range(max_N_simutation)):
        X_clone_list = []
        for t0 in time_info_order:
            t0_index = np.nonzero((adata.obs["time_info"] == t0).to_numpy())[0]
            # mask=np.zeros(X_clone.shape).astype(bool)
            # mask[t0_index]=True
            X_clone_t0 = X_clone[t0_index]
            np.random.shuffle(X_clone_t0)
            # np.putmask(X_clone,mask,X_clone_t0)
            X_clone_list.append(X_clone_t0)
        X_clone = np.vstack(X_clone_list)
        # print(X_clone.sum(0))
        # print(X_clone.sum(1))
        adata_rand.obsm["X_clone"] = ssp.csr_matrix(X_clone)

        fate_coupling(
            adata_rand,
            selected_fates=selected_fates,
            source="X_clone",
            normalize=normalize,
        )  # compute the fate coupling
        X_coupling_random.append(adata_rand.uns["fate_coupling_X_clone"]["X_coupling"])

    X_coupling_random = np.array(X_coupling_random)
    fate_names = adata.uns["fate_coupling_X_clone"]["fate_names"]

    settings.verbosity = old_verbosity

    pvalue = np.zeros((len(fate_names), len(fate_names)))
    pvalue_greater = np.zeros((len(fate_names), len(fate_names)))
    pvalue_less = np.zeros((len(fate_names), len(fate_names)))
    for i in range(len(fate_names)):
        for j in range(len(fate_names)):
            greater = (
                np.sum(np.array(X_coupling_random)[:, i, j] > X_coupling_true[i, j])
                / max_N_simutation
            )
            less = (
                np.sum(np.array(X_coupling_random)[:, i, j] < X_coupling_true[i, j])
                / max_N_simutation
            )
            pvalue_greater[i, j] = greater
            pvalue_less[i, j] = less
            pvalue[i, j] = np.min([greater, less])
            if i == j:
                pvalue[i, j] = np.nan
                pvalue_greater[i, j] = np.nan
                pvalue_less[i, j] = np.nan

    adata.uns["fate_coupling_X_clone"]["pvalue"] = pvalue
    adata.uns["fate_coupling_X_clone"]["pvalue_greater"] = pvalue_greater
    adata.uns["fate_coupling_X_clone"]["pvalue_less"] = pvalue_less
    logg.warn(f"pvalue updated at adata.uns['fate_coupling_X_clone']")

    return adata, X_coupling_random


def fate_map(
    adata,
    selected_fates=None,
    source="transition_map",
    map_backward=True,
    method="norm-sum",
    fate_count=False,
    force_run=False,
):
    """
    Compute transition probability to given fate/ancestor clusters.

    Given a transition map :math:`T_{ij}`, we explore build
    the fate map :math:`P_i(\mathcal{C})` towards a set of states annotated with
    fate :math:`\mathcal{C}` in the following ways.

    Step 1: Map normalization: :math:`T_{ij}\leftarrow T_{ij}/\sum_j T_{ij}`.

    Step 2: If `map_backward=False`, perform matrix transpose :math:`T_{ij} \leftarrow T_{ji}`.

    Step 3: aggregate fate probabiliteis within a given cluster :math:`\mathcal{C}`:

    * method='sum': :math:`P_i(\mathcal{C})=\sum_{j\in \mathcal{C}} T_{ij}`.
      This gives the intuitive meaning of fate probability.

    * method='norm-sum': We normalize the map from 'sum' method within a cluster, i.e.
      :math:`P_i(\mathcal{C})\leftarrow P_i(\mathcal{C})/\sum_j P_j(\mathcal{C})`.
      This gives the probability that a fate cluster :math:`\mathcal{C}` originates
      from an initial state :math:`i`.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    selected_fates: `list`, optional (default: all fates)
        List of cluster ids consistent with adata.obs['state_info'].
        It allows a nested list, where we merge clusters within
        each sub-list into a mega-fate cluster.
    source: `str`, optional (default: 'transition_map')
        The transition map to be used for plotting: {'transition_map',
        'intraclone_transition_map',...}. The actual available
        map depends on adata itself, which can be accessed at adata.uns['available_map']
    map_backward: `bool`, optional (default: True)
        If `map_backward=True`, show fate properties of initial cell states :math:`i`;
        otherwise, show progenitor properties of later cell states :math:`j`.
        This is used for building the fate map :math:`P_i(\mathcal{C})`. See :func:`.fate_map`.
    method: `str`, optional (default: 'norm-sum')
        Method to obtain the fate probability map :math:`P_i(\mathcal{C})` towards a set
        of states annotated with fate :math:`\mathcal{C}`. Available options:
        {'sum', 'norm-sum'}. See :func:`.fate_map`.
    fate_count: `bool`, optional (default: False)
        Used to determine the method for computing the fate potential of a state.
        If ture, just to count the number of possible fates; otherwise, use the Shannon entropy.
    force_run: `bool`, optional (default: False)
        Re-compute the fate map.

    Returns
    -------
    Fate map for each targeted fate cluster is updated at adata.obs[f'fate_map_{source}_{fate_name}'].
    The accompanying parameters are saved at adata.uns[f"fate_map_{source}_{fate}"]
    """

    hf.check_available_map(adata)

    if source not in adata.uns["available_map"]:
        raise ValueError(f"source should be among {adata.uns['available_map']}")

    else:

        state_annote = adata.obs["state_info"]
        (
            mega_cluster_list,
            __,
            __,
            sel_index_list,
        ) = hf.analyze_selected_fates(state_annote, selected_fates)

        key_word = f"fate_map_{source}"
        available_choices = hf.parse_output_choices(
            adata, key_word, where="obs", interrupt=False
        )

        # check if we need to recompute the map
        re_compute = True
        condi_0 = len(available_choices) > 0
        condi_1 = set(mega_cluster_list) <= set(available_choices)
        if condi_0 & condi_1:
            map_backward_all = set(
                [
                    adata.uns["fate_map_params"][f"{source}_{x}"]["map_backward"]
                    for x in mega_cluster_list
                ]
            )
            method_all = set(
                [
                    adata.uns["fate_map_params"][f"{source}_{x}"]["method"]
                    for x in mega_cluster_list
                ]
            )
            # check that the parameters are uniform and equals to input
            if len(map_backward_all) == 1 and len(method_all) == 1:
                condi_2 = map_backward == list(map_backward_all)[0]
                condi_3 = method == list(method_all)[0]
                if condi_2 and condi_3:
                    re_compute = False
        if not (re_compute or force_run):
            logg.info("Use pre-computed fate map")
        else:
            (
                fate_map,
                mega_cluster_list,
                relative_bias,
                expected_prob,
                valid_fate_list,
                sel_index_list,
                fate_entropy,
            ) = _utils.compute_fate_probability_map(
                adata,
                selected_fates=selected_fates,
                used_Tmap=source,
                map_backward=map_backward,
                method=method,
                fate_count=fate_count,
            )

            if map_backward:
                cell_id_t1 = adata.uns["Tmap_cell_id_t1"]
            else:
                cell_id_t1 = adata.uns["Tmap_cell_id_t2"]

            if "fate_map_params" not in adata.uns.keys():
                adata.uns[f"fate_map_params"] = {}

            if len(mega_cluster_list) == 0:
                logg.error("No cells selected. Computation aborted!")
            else:
                for j, fate in enumerate(mega_cluster_list):
                    temp_map = np.zeros(adata.shape[0]) + np.nan
                    temp_map[cell_id_t1] = fate_map[:, j]
                    adata.obs[f"fate_map_{source}_{fate}"] = temp_map
                    adata.uns[f"fate_map_params"][f"{source}_{fate}"] = {
                        "map_backward": map_backward,
                        "method": method,
                    }
                    logg.info(f"Results saved at adata.obs['fate_map_{source}_{fate}']")

                temp_map = np.zeros(adata.shape[0]) + np.nan
                temp_map[cell_id_t1] = fate_entropy
                adata.uns[f"fate_potency_tmp"] = temp_map


@_doc_params(
    selected_fates=_docs.selected_fates,
    all_source=_docs.all_source,
    map_source=_docs.map_source,
    map_backward=_docs.map_backward,
    selected_times=_docs.selected_times,
    fate_method=_docs.fate_method,
)
def fate_potency(
    adata,
    selected_fates=None,
    source="transition_map",
    map_backward=True,
    method="norm-sum",
    fate_count=False,
):
    """
    Quantify how multi-potent a cell state is.

    It runs :func:`.fate_map` to compute the fate potency. Please see all parameter definitions there.

    Parameters
    ----------
    {selected_fates}
    {map_source}
    {map_backward}
    {fate_method}
    fate_count:
        True: count the number of possible fates; otherwise, use the Shannon entropy.

    Returns
    -------
    Results saved at adata.obs[f"fate_potency_{{source}}"].
    """

    fate_map(
        adata,
        selected_fates=selected_fates,
        source=source,
        map_backward=map_backward,
        method=method,
        fate_count=fate_count,
        force_run=True,
    )
    adata.obs[f"fate_potency_{source}"] = adata.uns["fate_potency_tmp"]
    if "fate_potency_params" not in adata.uns.keys():
        adata.uns[f"fate_potency_params"] = {}
    adata.uns[f"fate_potency_params"][f"{source}"] = {
        "map_backward": map_backward,
        "method": method,
    }
    logg.info(f"Results saved at adata.obs['fate_potency_{source}']")


def fate_bias(
    adata,
    selected_fates=None,
    source="transition_map",
    map_backward=True,
    method="norm-sum",
    sum_fate_prob_thresh=0,
    pseudo_count=0,
):
    """
    Compute fate bias to given two fate clusters (A, B).

    Given a fate map :math:`P_i` towards two fate clusters
    :math:`\{\mathcal{A}, \mathcal{B}\}`, constructed according
    to :func:`.fate_map`, we compute the fate bias of state :math:`i` as
    :math:`[P(\mathcal{A})+c_0]/[P(\mathcal{A})+P(\mathcal{B})+2c_0]`,
    where :math:`c_0=a * \max_{i,\mathcal{C}} P_i(\mathcal{C})`
    is a re-scaled pseudocount, with :math:`a` given by pseudo_count.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    selected_fates: `list`
        List of cluster ids consistent with adata.obs['state_info'].
        It allows a nested structure. If so, we merge clusters within
        each sub-list into a mega-fate cluster.
    source: `str`
        The transition map to be used for plotting: {'transition_map',
        'intraclone_transition_map',...}. The actual available
        map depends on adata itself, which can be accessed at adata.uns['available_map']
    map_backward: `bool`, optional (default: True)
        If `map_backward=True`, show fate properties of initial cell states :math:`i`;
        otherwise, show progenitor properties of later cell states :math:`j`.
        This is used for building the fate map :math:`P_i(\mathcal{C})`. See :func:`.fate_map`.
    method: `str`, optional (default: 'norm-sum')
        Method to obtain the fate probability map :math:`P_i(\mathcal{C})` towards a set
        of states annotated with fate :math:`\mathcal{C}`. Available options:
        {'sum', 'norm-sum'}. See :func:`.fate_map`.
    sum_fate_prob_thresh: `float`, optional (default: 0.05)
        The fate bias of a state is plotted only when it has a cumulative fate
        probability to the combined cluster (A+B) larger than this threshold,
        i.e., P(i->A)+P(i+>B) >  sum_fate_prob_thresh.

    Returns
    -------
    Results updated at adata.obs[f'fate_bias_{source}_{fate_1}_{fate_2}']
    """

    state_annote = adata.obs["state_info"]
    (
        mega_cluster_list,
        __,
        __,
        sel_index_list,
    ) = hf.analyze_selected_fates(state_annote, selected_fates)

    if len(mega_cluster_list) != 2:
        raise ValueError("selected_fates must have only two valid fates")

    fate_map(
        adata,
        selected_fates=selected_fates,
        source=source,
        map_backward=map_backward,
        method=method,
        force_run=True,
    )

    if map_backward:
        cell_id_t1 = adata.uns["Tmap_cell_id_t1"]
    else:
        cell_id_t1 = adata.uns["Tmap_cell_id_t2"]

    fate_vector_1 = np.array(adata.obs[f"fate_map_{source}_{mega_cluster_list[0]}"])[
        cell_id_t1
    ]
    fate_vector_2 = np.array(adata.obs[f"fate_map_{source}_{mega_cluster_list[1]}"])[
        cell_id_t1
    ]

    if pseudo_count == 0:
        pseudo_count = 10 ** (-10)

    add_count = pseudo_count * np.max([fate_vector_1, fate_vector_2])
    fate_vector_1 = fate_vector_1 + add_count
    fate_vector_2 = fate_vector_2 + add_count

    tot_prob = fate_vector_1 + fate_vector_2
    valid_idx = tot_prob > sum_fate_prob_thresh  # default 0.05
    fate_bias_vector = fate_vector_1[valid_idx] / (tot_prob[valid_idx])

    temp_map = np.zeros(adata.shape[0]) + 0.5  # initialize to be neutral
    temp_map[cell_id_t1[valid_idx]] = fate_bias_vector
    adata.obs[
        f"fate_bias_{source}_{mega_cluster_list[0]}*{mega_cluster_list[1]}"
    ] = temp_map

    if "fate_bias_params" not in adata.uns.keys():
        adata.uns[f"fate_bias_params"] = {}
    adata.uns[f"fate_bias_params"][
        f"{source}_{mega_cluster_list[0]}*{mega_cluster_list[1]}"
    ] = {"map_backward": map_backward, "method": method}

    logg.info(
        f"Results saved at adata.obs['fate_bias_{source}_{mega_cluster_list[0]}*{mega_cluster_list[1]}']"
    )


def progenitor(
    adata,
    selected_fates=None,
    source="transition_map",
    map_backward=True,
    method="norm-sum",
    bias_threshold_A=0.5,
    bias_threshold_B=0.5,
    sum_fate_prob_thresh=0,
    pseudo_count=0,
    avoid_target_states=False,
):
    """
    Identify trajectories towards/from two given clusters.

    Given fate bias :math:`Q_i` for a state :math:`i` as defined in :func:`.fate_bias`,
    the selected ancestor population satisfies:

       * :math:`P_i(\mathcal{A})+P_i(\mathcal{B})` > sum_fate_prob_thresh;

       * Ancestor population for fate :math:`\mathcal{A}` satisfies :math:`Q_i` > bias_threshold_A

       * Ancestor population for fate :math:`\mathcal{B}` satisfies :math:`Q_i` < bias_threshold_B

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    selected_fates: `list`
        List of cluster ids consistent with adata.obs['state_info'].
        It allows a nested structure.
    source: `str`
        The transition map to be used for plotting: {'transition_map',
        'intraclone_transition_map',...}. The actual available
        map depends on adata itself, which can be accessed at adata.uns['available_map']
    map_backward: `bool`, optional (default: True)
        If `map_backward=True`, show fate properties of initial cell states :math:`i`;
        otherwise, show progenitor properties of later cell states :math:`j`.
        This is used for building the fate map :math:`P_i(\mathcal{C})`. See :func:`.fate_map`.
    method: `str`, optional (default: 'norm-sum')
        Method to obtain the fate probability map :math:`P_i(\mathcal{C})` towards a set
        of states annotated with fate :math:`\mathcal{C}`. Available options:
        {'sum', 'norm-sum'}. See :func:`.fate_map`.
    bias_threshold_A: `float`, optional (default: 0), range: (0,1)
        The threshold for selecting ancestor population for fate A.
    bias_threshold_B: `float`, optional (default: 0), range: (0,1)
        The threshold for selecting ancestor population for fate B.
    sum_fate_prob_thresh: `float`, optional (default: 0), range: (0,1)
        Minimum cumulative probability towards joint cluster (A,B)
        to qualify for ancestor selection.
    pseudo_count: `float`, optional (default: 0)
        Pseudo count to compute the fate bias. The bias = (Pa+c0)/(Pa+Pb+2*c0),
        where c0=pseudo_count*(maximum fate probability) is a rescaled pseudo count.
    avoid_target_states:
        If True, exclude target states for computing fate bias.

    Returns
    -------
    Results saved at adata.obs[f'progenitor_{source}_{fate_name}'] and adata.obs[f'diff_trajectory_{source}_{fate_name}']"
    """

    state_info = np.array(adata.obs["state_info"])
    (
        mega_cluster_list,
        valid_fate_list,
        __,
        sel_index_list,
    ) = hf.analyze_selected_fates(state_info, selected_fates)

    if len(mega_cluster_list) != 2:
        raise ValueError("selected_fates must have only two valid fates")

    fate_bias(
        adata,
        selected_fates=selected_fates,
        source=source,
        map_backward=map_backward,
        method=method,
        sum_fate_prob_thresh=sum_fate_prob_thresh,
        pseudo_count=pseudo_count,
    )

    if map_backward:
        cell_id_t1 = adata.uns["Tmap_cell_id_t1"]
    else:
        cell_id_t1 = adata.uns["Tmap_cell_id_t2"]

    fate_key = f"{mega_cluster_list[0]}*{mega_cluster_list[1]}"
    fate_vector = np.array(adata.obs[f"fate_bias_{source}_{fate_key}"])[cell_id_t1]

    idx_for_group_A = fate_vector > (bias_threshold_A)
    idx_for_group_B = fate_vector < (bias_threshold_B)

    if avoid_target_states:
        for zz in valid_fate_list[0]:
            id_A_t1 = np.nonzero(state_info[cell_id_t1] == zz)[0]
            idx_for_group_A[id_A_t1] = False

        for zz in valid_fate_list[1]:
            id_B_t1 = np.nonzero(state_info[cell_id_t1] == zz)[0]
            idx_for_group_B[id_B_t1] = False

    group_A_idx_full = np.zeros(adata.shape[0], dtype=bool)
    group_A_idx_full[cell_id_t1] = idx_for_group_A
    group_B_idx_full = np.zeros(adata.shape[0], dtype=bool)
    group_B_idx_full[cell_id_t1] = idx_for_group_B

    # store the trajectory
    key_word = "progenitor"
    if f"{key_word}_params" not in adata.uns.keys():
        adata.uns[f"{key_word}_params"] = {}

    temp_list = [group_A_idx_full, group_B_idx_full]
    for j, fate_name in enumerate(mega_cluster_list):
        adata.obs[f"{key_word}_{source}_{fate_name}"] = temp_list[j]
        selected_idx = sel_index_list[j]
        combined_prob_temp = temp_list[j].astype(int) + selected_idx.astype(int)
        adata.obs[f"diff_trajectory_{source}_{fate_name}"] = combined_prob_temp
        adata.obs[f"progenitor_{source}_{fate_name}"] = temp_list[j].astype(int)
        adata.uns[f"{key_word}_params"][f"{source}_{fate_name}"] = {
            "map_backward": map_backward,
            "method": method,
        }
        logg.info(
            f"Results saved at adata.obs[f'progenitor_{source}_{fate_name}'] and adata.obs[f'diff_trajectory_{source}_{fate_name}']"
        )


def iterative_differentiation(
    adata,
    selected_fates=None,
    source="transition_map",
    map_backward=True,
    map_threshold=0.1,
    apply_time_constaint=False,
):
    """
    Infer trajectory towards/from a cluster

    If map_backward=True, infer the trajectory backward in time.
    Using inferred transition map, the inference is applied recursively. It
    starts with the cell states for the selected fate and uses the selected
    map to infer the immediate ancestor states. Then, using these putative
    ancestor states as the secondary input, it finds their immediate ancestors
    again. This goes on until all time points are exhausted.

    It only works for transition map from multi-time clones.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    selected_fate: `str`, or `list`
        Targeted cluster of the trajectory, as consistent with adata.obs['state_info']
        When it is a list, the listed clusters are combined into a single fate cluster.
    used_Tmap: `str`
        The transition map to be used for plotting: {'transition_map',
        'intraclone_transition_map',...}. The actual available
        map depends on adata itself, which can be accessed at adata.uns['available_map']
    map_backward: `bool`, optional (default: True)
        If `map_backward=True`, show fate properties of initial cell states :math:`i`;
        otherwise, show progenitor properties of later cell states :math:`j`.
        This is used for building the fate map :math:`P_i(\mathcal{C})`. See :func:`.fate_map`.
    map_threshold: `float`, optional (default: 0.1)
        Relative threshold in the range [0,1] for truncating the fate map
        towards the cluster. Only states above the threshold will be selected.
    apply_time_constaint: `bool`, optional (default: False)
        If true, in each iteration of finding the immediate ancestor states, select cell states
        at the corresponding time point.

    Returns
    -------
    adata.obs[f'traj_{source}_{fate_name}']: `np.array`
        The probability of each state to belong to a trajectory.
    """

    # We always use the probabilistic map, which is more realiable. Otherwise, the result is very sensitive to thresholding
    hf.check_available_map(adata)
    if source not in adata.uns["available_map"]:
        raise ValueError(f"source should be among {adata.uns['available_map']}")
    else:
        # we normalize the map in advance to avoid normalization later in mapout_trajectories
        used_map = adata.uns[source]
        resol = 10 ** (-10)
        used_map = hf.sparse_rowwise_multiply(
            used_map, 1 / (resol + np.sum(used_map, 1).A.flatten())
        )

        if not map_backward:
            used_map = used_map.T

        state_annote = adata.obs["state_info"]
        (
            mega_cluster_list,
            __,
            __,
            sel_index_list,
        ) = hf.analyze_selected_fates(state_annote, selected_fates)

        if len(mega_cluster_list) == 0:
            logg.error("No cells selected. Computation aborted!")
        else:
            for k0, fate_key in enumerate(mega_cluster_list):
                selected_idx = sel_index_list[k0]

                time_info = np.array(adata.obs["time_info"])
                if map_backward:
                    cell_id_t1 = adata.uns["Tmap_cell_id_t1"]
                    cell_id_t2 = adata.uns["Tmap_cell_id_t2"]
                    sort_time_info = np.sort(list(set(time_info)))[::-1]
                else:
                    cell_id_t2 = adata.uns["Tmap_cell_id_t1"]
                    cell_id_t1 = adata.uns["Tmap_cell_id_t2"]
                    sort_time_info = np.sort(list(set(time_info)))

                if apply_time_constaint:
                    selected_idx = selected_idx * (time_info == sort_time_info[0])
                prob_array = [selected_idx]

                # used_map=hf.sparse_column_multiply(used_map,1/(resol+used_map.sum(0)))
                for j, t_0 in enumerate(sort_time_info[1:]):
                    prob_1r_full = np.zeros(adata.shape[0])
                    prob_1r_full[cell_id_t1] = _utils.mapout_trajectories(
                        used_map,
                        prob_array[j],
                        threshold=map_threshold,
                        cell_id_t1=cell_id_t1,
                        cell_id_t2=cell_id_t2,
                    )

                    ## thresholding the result
                    prob_1r_full = prob_1r_full * (
                        prob_1r_full > map_threshold * np.max(prob_1r_full)
                    )

                    if apply_time_constaint:
                        prob_1r_full = prob_1r_full * (time_info == t_0)

                    # rescale
                    prob_1r_full = prob_1r_full / (np.max(prob_1r_full) + resol)

                    prob_array.append(prob_1r_full)

                cumu_prob = np.array(prob_array).sum(0)
                key_word = f"iterative_diff_{source}_{fate_key}"
                adata.uns[key_word] = {
                    "diff_prob_list": prob_array,
                    "sorted_time_info": sort_time_info,
                    "map_backward": map_backward,
                }
                adata.obs[f"diff_trajectory_{source}_{fate_key}"] = cumu_prob
                logg.info(
                    f"Results saved at adata.obs[f'diff_trajectory_{source}_{fate_key}']"
                )



"""KAILINss - dynamic inference by integrating transcriptome and lineage information"""

#计算稀疏矩阵matrix
#如果输入draw，就转变成非稀疏矩阵并开始画图

def merge_tissue_with_singlecell(
    current_folder,
    h5ad_filename,
    csv_filename,
    sample_choose,
    num_samples,
    ):
        import numpy as np
        import os
        import kailin as kl
        current_folder_filter = os.path.join(current_folder, sample_choose)
        current_folder_filter_1 = os.path.join(current_folder_filter, 'data')
        filter_sample = os.path.join(current_folder_filter_1, h5ad_filename)


        csv_sample = os.path.join(current_folder_filter_1, csv_filename)

        import scanpy as sc
        import anndata as ad
        # 假设 filter_sample 是 AnnData 文件路径
        adata_path = filter_sample
        # 使用 anndata 读取文件
        adata_singlecell = ad.read_h5ad(adata_path)

        row_names_cell = adata_singlecell.obs_names

        # 读取列名（基因名）
        column_names_gene = adata_singlecell.var_names
        
        import pandas as pd
        csv_data = pd.read_csv(csv_sample)
        print(csv_data)

        #读取csv_data有多少列
        num_columns = csv_data.shape[1]

        # 将 gene_name 列存储到变量 tissue_gene 中
        tissue_gene = csv_data['gene_name']


        # 确保两者都是集合（去重处理）
        common_elements = set(column_names_gene).intersection(set(tissue_gene))

        # 计算共有元素数量
        num_common_elements = len(common_elements)


        # 转换为列表（如果是 Pandas Series）
        singecell_gene = list(column_names_gene)
        tissue_gene = list(tissue_gene)

        # 找出共有元素
        common_elements = set(singecell_gene).intersection(set(tissue_gene))

        # 找出共有元素在 column_names_gene 和 tissue_gene 中的行号
        singlecell_indices = [i for i, gene in enumerate(singecell_gene) if gene in common_elements]
        tissue_indices = [i for i, gene in enumerate(tissue_gene) if gene in common_elements]


        #过滤adata
        adata_singlecell_filter = adata_singlecell[:,singlecell_indices]


        #过滤tissue
        data_tissue_filter = csv_data.iloc[tissue_indices, :]


        # 将基因名设置为索引，方便排序
        data_tissue_filter = data_tissue_filter.set_index('gene_name')

        # 按照 adata_singlecell.var_names 的顺序重新排序
        data_tissue_filter = data_tissue_filter.reindex(adata_singlecell_filter.var_names)

        data_tissue_filter_t = data_tissue_filter.T

        import scipy.sparse as sp

        adata_singlecell_filter_to = adata_singlecell_filter.X.toarray()

        #合并表达矩阵
        from scipy.sparse import csr_matrix
        merged_X = np.vstack((adata_singlecell_filter_to, data_tissue_filter_t.values))
        merged_X = csr_matrix(merged_X)

        #合并var cell_name
        merged_varname = adata_singlecell_filter.var_names


        #存储obs gene_name
        #merged_varname = [adata_singlecell_filter.var_names,data_tissue_filter_t.index]
        merged_obsname = pd.Index(adata_singlecell_filter.obs_names).append(data_tissue_filter_t.index)

        import pandas as pd
        import anndata as ad
        merged_obsname_df = pd.DataFrame(index=merged_obsname, columns=[])  # 空 DataFrame 仅带索引
        merged_varname_df = pd.DataFrame(index=merged_varname, columns=[])

        # 创建 AnnData 对象
        adata_merged = ad.AnnData(X=merged_X, obs=merged_obsname_df, var=merged_varname_df)

        print(adata_merged)

        return adata_merged
    







import os

import numpy as np
import scipy.sparse as ssp
from matplotlib import pyplot as plt

from cospar.plotting import _utils as pl_util

from .. import help_functions as hf
from .. import logging as logg
from .. import settings
from ..help_functions import _docs
from ..help_functions._docs import _doc_params

color_map_reds = pl_util.darken_cmap(plt.cm.Reds, scale_factor=0.9)
color_map_coolwarm = pl_util.darken_cmap(plt.cm.coolwarm, scale_factor=1)


@_doc_params(
    selected_fates=_docs.selected_fates,
    source=_docs.all_source,
    color_bar=_docs.color_bar,
    rename_fates=_docs.rename_fates,
    color_map=_docs.color_map,
    figure_index=_docs.figure_index,
)
def fate_coupling(
    adata,
    source="transition_map",
    color_bar=True,
    rename_fates=None,
    color_map=color_map_reds,
    figure_index="",
    color_bar_label="Clonal coupling score",
    title=None,
    vmax=None,
    **kwargs,
):
    """
    Plot fate coupling determined by the transition map.

    The results should be pre-computed from :func:`cospar.tl.fate_coupling`

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    {source}
    {color_bar}
    {rename_fates}
    {color_map}
    {figure_index}

    Returns
    -------
    ax:
        The axis object of this plot.
    """

    fig_width = settings.fig_width
    fig_height = settings.fig_height
    figure_path = settings.figure_path

    data_des = adata.uns["data_des"][-1]

    if color_bar:
        fig_width = fig_width + 0.5

    key_word = f"fate_coupling_{source}"
    available_choices = hf.parse_output_choices(adata, key_word, where="uns")
    X_coupling = adata.uns[key_word]["X_coupling"]
    fate_names = adata.uns[key_word]["fate_names"]
    rename_fates = hf.rename_list(fate_names, rename_fates)

    if "x_ticks" not in kwargs.keys():
        kwargs["x_ticks"] = rename_fates
    if "y_ticks" not in kwargs.keys():
        kwargs["y_ticks"] = rename_fates

    if vmax is None:
        # vmax = (
        #     np.percentile(X_coupling - np.diag(np.diag(X_coupling)), 95)
        #     + np.percentile(X_coupling - np.diag(np.diag(X_coupling)), 98)
        # ) / 2
        vmax = np.max(X_coupling - np.diag(np.diag(X_coupling)))

    ax = pl_util.heatmap(
        X_coupling,
        color_bar_label=color_bar_label,
        color_bar=color_bar,
        fig_width=fig_width,
        fig_height=fig_height,
        color_map=color_map,
        vmax=vmax,
        **kwargs,
    )

    plt.tight_layout()
    if title is None:
        ax.set_title(f"source: {source}")
    else:
        ax.set_title(title)

    if figure_index != "":
        figure_index == f"_{figure_index}"
    plt.savefig(
        os.path.join(
            figure_path,
            f"{data_des}_{key_word}{figure_index}.{settings.file_format_figs}",
        )
    )
    return ax


@_doc_params(
    source=_docs.all_source,
    rename_fates=_docs.rename_fates,
    color_map=_docs.color_map,
)
def fate_hierarchy(
    adata,
    source="transition_map",
    rename_fates=None,
    plot_history=False,
):
    """
    Plot fate hierarchy determined by the transition map.

    The results should be pre-computed from :func:`cospar.tl.fate_hierarchy`.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    {source}
    {rename_fates}
    plot_history:
        True: plot the history of iterations.

    Returns
    -------
    ax:
        The axis object of this plot.
    """

    key_word = f"fate_hierarchy_{source}"
    available_choices = hf.parse_output_choices(
        adata, key_word, where="uns"
    )  # it also performs checks
    node_mapping = adata.uns[key_word]["node_mapping"]
    history = adata.uns[key_word]["history"]
    fate_names = adata.uns[key_word]["fate_names"]
    fate_tree = adata.uns[key_word]["tree"]
    rename_fates = hf.rename_list(fate_names, rename_fates)

    # t = pl_util.convert_to_tree(parent_map, rename_fates)
    print(fate_tree)
    if plot_history:
        pl_util.plot_neighbor_joining(
            settings.figure_path,
            node_mapping,
            history[0],
            history[1],
            history[2],
        )


@_doc_params(
    selected_fates=_docs.selected_fates,
    source=_docs.map_source,
    selected_times=_docs.selected_times,
    background=_docs.background,
    show_histogram=_docs.show_histogram,
    color_bar=_docs.color_bar,
    rename_fates=_docs.rename_fates,
    color_map=_docs.color_map,
    mask=_docs.mask,
    target_transparency=_docs.target_transparency,
    figure_index=_docs.figure_index,
)
def fate_map(
    adata,
    selected_fates=None,
    source="transition_map",
    selected_times=None,
    background=True,
    show_histogram=False,
    plot_target_state=False,
    auto_color_scale=True,
    color_bar=True,
    target_transparency=0.2,
    figure_index="",
    mask=None,
    color_map=color_map_reds,
    **kwargs,
):
    """
    Plot transition probability to given fate/ancestor clusters.

    The results should be pre-computed from :func:`cospar.tl.fate_map`.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    {selected_fates}
    {source}
    {selected_times}
    {background}
    {show_histogram}
    plot_target_state:
        If True, plot target states.
    {color_bar}
    auto_color_scale:
        True: automatically rescale the color range to match the value range.
    {target_transparency}
    {figure_index}
    {mask}
    {color_map}
    """

    key_word = f"fate_map_{source}"
    available_choices = hf.parse_output_choices(adata, key_word, where="obs")
    time_info = np.array(adata.obs["time_info"])
    state_info = adata.obs["state_info"]
    if selected_fates is None:
        selected_fates = sorted(list(set(state_info)))
    (
        mega_cluster_list,
        valid_fate_list,
        __,
        sel_index_list,
    ) = hf.analyze_selected_fates(state_info, selected_fates)

    for j, fate_key in enumerate(mega_cluster_list):
        if fate_key not in available_choices:
            logg.error(
                f"The fate map for {fate_key} have not been computed yet. Skipped!"
            )
        else:
            fate_vector = np.array(adata.obs[f"fate_map_{source}_{fate_key}"])
            params = adata.uns["fate_map_params"][f"{source}_{fate_key}"]
            map_backward = params["map_backward"]
            method = params["method"]

            if map_backward:
                cell_id_t1 = adata.uns["Tmap_cell_id_t1"]
            else:
                cell_id_t1 = adata.uns["Tmap_cell_id_t2"]
            sp_idx = hf.selecting_cells_by_time_points(
                time_info[cell_id_t1], selected_times
            )

            if method == "norm-sum":
                color_bar_label = "Progenitor prob."
            else:
                color_bar_label = "Fate prob."

            if plot_target_state:
                target_list = valid_fate_list[j]
            else:
                target_list = None

            pl_util.fate_map_embedding(
                adata,
                fate_vector,
                cell_id_t1,
                sp_idx,
                mask=mask,
                target_list=target_list,
                color_bar_label=color_bar_label,
                color_bar_title="",
                figure_title=fate_key,
                background=background,
                show_histogram=show_histogram,
                auto_color_scale=auto_color_scale,
                color_bar=color_bar,
                target_transparency=target_transparency,
                color_map=color_map,
                **kwargs,
            )

            plt.tight_layout()
            data_des = adata.uns["data_des"][-1]
            if figure_index != "":
                figure_index == f"_{figure_index}"
            plt.savefig(
                os.path.join(
                    f"{settings.figure_path}",
                    f"{data_des}_{key_word}_{fate_key}{figure_index}.{settings.file_format_figs}",
                )
            )


@_doc_params(
    source=_docs.map_source,
    selected_times=_docs.selected_times,
    background=_docs.background,
    show_histogram=_docs.show_histogram,
    color_bar=_docs.color_bar,
    rename_fates=_docs.rename_fates,
    color_map=_docs.color_map,
    mask=_docs.mask,
    target_transparency=_docs.target_transparency,
    figure_index=_docs.figure_index,
)
def fate_potency(
    adata,
    source="transition_map",
    selected_times=None,
    background=True,
    show_histogram=False,
    auto_color_scale=True,
    color_bar=True,
    figure_index="",
    mask=None,
    color_map=color_map_reds,
    **kwargs,
):
    """
    Plot fate potency.

    The results should be pre-computed from :func:`cospar.tl.fate_potency`.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    {source}
    {selected_times}
    {background}
    {show_histogram}
    {color_bar}
    auto_color_scale:
        True: automatically rescale the color range to match the value range.
    {figure_index}
    {mask}
    {color_map}
    """

    key_word_0 = "fate_potency"
    key_word = f"{key_word_0}_{source}"
    available_choices = hf.parse_output_choices(adata, key_word, where="obs")
    map_backward = adata.uns[f"{key_word_0}_params"][f"{source}"]["map_backward"]
    if map_backward:
        cell_id_t1 = adata.uns["Tmap_cell_id_t1"]
    else:
        cell_id_t1 = adata.uns["Tmap_cell_id_t2"]

    time_info = np.array(adata.obs["time_info"])
    sp_idx = hf.selecting_cells_by_time_points(time_info[cell_id_t1], selected_times)

    fate_vector = np.array(adata.obs[key_word])

    pl_util.fate_map_embedding(
        adata,
        fate_vector,
        cell_id_t1,
        sp_idx,
        mask=mask,
        target_list=None,
        color_bar_label="Fate potency",
        color_bar_title="",
        figure_title="",
        background=background,
        show_histogram=show_histogram,
        auto_color_scale=auto_color_scale,
        color_bar=color_bar,
        color_map=color_map,
        **kwargs,
    )

    plt.tight_layout()
    data_des = adata.uns["data_des"][-1]
    if figure_index != "":
        figure_index == f"_{figure_index}"

    plt.savefig(
        os.path.join(
            f"{settings.figure_path}",
            f"{data_des}_{key_word}{figure_index}.{settings.file_format_figs}",
        )
    )


@_doc_params(
    selected_fates=_docs.selected_fates,
    source=_docs.map_source,
    selected_times=_docs.selected_times,
    background=_docs.background,
    show_histogram=_docs.show_histogram,
    color_bar=_docs.color_bar,
    rename_fates=_docs.rename_fates,
    color_map=_docs.color_map,
    mask=_docs.mask,
    target_transparency=_docs.target_transparency,
    figure_index=_docs.figure_index,
)
def fate_bias(
    adata,
    selected_fates=None,
    source="transition_map",
    selected_times=None,
    background=True,
    show_histogram=False,
    plot_target_state=False,
    auto_color_scale=False,
    color_bar=True,
    target_transparency=0.2,
    figure_index="",
    horizontal=False,
    mask=None,
    color_bar_title=None,
    color_map=color_map_coolwarm,
    **kwargs,
):
    """
    Plot fate bias.

    The results should be pre-computed from :func:`cospar.tl.fate_bias`.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    {selected_fates}
    {source}
    {selected_times}
    {background}
    {show_histogram}
    plot_target_state:
        If True, plot target states.
    {color_bar}
    auto_color_scale:
        True: automatically rescale the color range to match the value range.
    {target_transparency}
    {figure_index}
    horizontal: `bool`
        If true, display figures horizontally.
    {mask}
    color_bar_title: `str`
        Title for color bar.
    {color_map}
    """

    key_word_0 = "fate_bias"
    key_word = f"{key_word_0}_{source}"
    available_choices = hf.parse_output_choices(adata, key_word, where="obs")

    if selected_fates is None:
        logg.info(
            "selected_fates not specified. Using the first available pre-computed fate_bias"
        )
        fate_key = available_choices[0]
        valid_fate_list = [xx.split("_") for xx in fate_key.split("*")]
    else:
        state_info = adata.obs["state_info"]
        (
            mega_cluster_list,
            valid_fate_list,
            __,
            sel_index_list,
        ) = hf.analyze_selected_fates(state_info, selected_fates)

        if len(mega_cluster_list) != 2:
            raise ValueError("selected_fates must have only two valid fates")

        fate_key = f"{mega_cluster_list[0]}*{mega_cluster_list[1]}"

    params = adata.uns[f"{key_word_0}_params"][f"{source}_{fate_key}"]
    map_backward = params["map_backward"]
    method = params["method"]

    if map_backward:
        cell_id_t1 = adata.uns["Tmap_cell_id_t1"]
    else:
        cell_id_t1 = adata.uns["Tmap_cell_id_t2"]
    time_info = np.array(adata.obs["time_info"])
    sp_idx = hf.selecting_cells_by_time_points(time_info[cell_id_t1], selected_times)

    fate_vector = np.array(adata.obs[f"fate_bias_{source}_{fate_key}"])

    if plot_target_state:
        target_list = valid_fate_list[0] + valid_fate_list[1]
    else:
        target_list = None

    if color_bar_title is None:
        color_bar_title = fate_key.split("*")[0]

    if method == "norm-sum":
        color_bar_label = "Progenitor bias"
    else:
        color_bar_label = "Fate bias"

    pl_util.fate_map_embedding(
        adata,
        fate_vector,
        cell_id_t1,
        sp_idx,
        mask=mask,
        color_map=color_map,
        target_list=target_list,
        color_bar_label=color_bar_label,
        color_bar_title=color_bar_title,
        figure_title="",
        background=background,
        show_histogram=show_histogram,
        auto_color_scale=auto_color_scale,
        color_bar=color_bar,
        horizontal=horizontal,
        target_transparency=target_transparency,
        histogram_scales=[0, 1],
        order_method="fate_bias",
        **kwargs,
    )

    plt.tight_layout()
    data_des = adata.uns["data_des"][-1]
    if figure_index != "":
        figure_index == f"_{figure_index}"

    plt.savefig(
        os.path.join(
            f"{settings.figure_path}",
            f"{data_des}_{key_word}_{fate_key}{figure_index}.{settings.file_format_figs}",
        )
    )


@_doc_params(
    selected_fates=_docs.selected_fates,
    source=_docs.map_source,
    selected_times=_docs.selected_times,
    background=_docs.background,
    show_histogram=_docs.show_histogram,
    color_bar=_docs.color_bar,
    rename_fates=_docs.rename_fates,
    color_map=_docs.color_map,
    mask=_docs.mask,
    target_transparency=_docs.target_transparency,
    figure_index=_docs.figure_index,
)
def progenitor(
    adata,
    selected_fates=None,
    source="transition_map",
    selected_times=None,
    background=True,
    plot_target_state=False,
    auto_color_scale=False,
    target_transparency=0.2,
    figure_index="",
    mask=None,
    **kwargs,
):
    """
    Plot the progenitors of given fate clusters.

    The results should be pre-computed from :func:`cospar.tl.progenitor`.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    {selected_fates}
    {source}
    {selected_times}
    {background}
    plot_target_state:
        If True, plot target states.
    auto_color_scale:
        True: automatically rescale the color range to match the value range.
    {target_transparency}
    {figure_index}
    {mask}
    """

    key_word_0 = "progenitor"
    key_word = f"{key_word_0}_{source}"
    available_choices = hf.parse_output_choices(adata, key_word, where="obs")

    if selected_fates is None:
        mega_cluster_list = available_choices
        valid_fate_list = [xx.split("_") for xx in mega_cluster_list]
    else:
        state_info = adata.obs["state_info"]
        (
            mega_cluster_list,
            valid_fate_list,
            __,
            sel_index_list,
        ) = hf.analyze_selected_fates(state_info, selected_fates)

    for j, fate_key in enumerate(mega_cluster_list):
        if fate_key not in available_choices:
            raise ValueError(f"The {key_word_0} has not been computed for {fate_key}.")

        params = adata.uns[f"{key_word_0}_params"][f"{source}_{fate_key}"]
        map_backward = params["map_backward"]

        if map_backward:
            cell_id_t1 = adata.uns["Tmap_cell_id_t1"]
        else:
            cell_id_t1 = adata.uns["Tmap_cell_id_t2"]
        time_info = np.array(adata.obs["time_info"])
        sp_idx = hf.selecting_cells_by_time_points(
            time_info[cell_id_t1], selected_times
        )

        fate_vector = np.array(adata.obs[f"{key_word_0}_{source}_{fate_key}"])

        if plot_target_state:
            target_list = valid_fate_list[j]
        else:
            target_list = None

        pl_util.fate_map_embedding(
            adata,
            fate_vector,
            cell_id_t1,
            sp_idx,
            mask=mask,
            target_list=target_list,
            color_bar_label="",
            color_bar_title="",
            figure_title="",
            background=background,
            show_histogram=False,
            auto_color_scale=auto_color_scale,
            color_bar=False,
            target_transparency=target_transparency,
            **kwargs,
        )

        plt.tight_layout()
        data_des = adata.uns["data_des"][-1]
        if figure_index != "":
            figure_index == f"_{figure_index}"

        plt.savefig(
            os.path.join(
                f"{settings.figure_path}",
                f"{data_des}_{key_word}_{fate_key}{figure_index}.{settings.file_format_figs}",
            )
        )


@_doc_params(
    selected_fates=_docs.selected_fates,
    source=_docs.map_source,
    figure_index=_docs.figure_index,
)
def iterative_differentiation(
    adata,
    selected_fates=None,
    source="transition_map",
    figure_index="",
):
    """
    Plot the progenitors of given fates across time points

    The results should be pre-computed from :func:`cospar.tl.iterative_differentiation`.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    {selected_fates}
    {source}
    {figure_index}
    """
    key_word_0 = "iterative_diff"
    key_word = f"{key_word_0}_{source}"
    available_choices = hf.parse_output_choices(adata, key_word, where="uns")

    if selected_fates is None:
        mega_cluster_list = available_choices
    else:
        state_info = adata.obs["state_info"]
        (
            mega_cluster_list,
            valid_fate_list,
            __,
            sel_index_list,
        ) = hf.analyze_selected_fates(state_info, selected_fates)

    x_emb = adata.obsm["X_emb"][:, 0]
    y_emb = adata.obsm["X_emb"][:, 1]
    point_size = settings.fig_point_size
    fig_width = settings.fig_width
    fig_height = settings.fig_height
    for j, fate_key in enumerate(mega_cluster_list):
        if fate_key not in available_choices:
            raise ValueError(f"The diff. traj. has not been computed for {fate_key}.")

        prob_array = np.array(
            adata.uns[f"{key_word_0}_{source}_{fate_key}"]["diff_prob_list"]
        )
        sorted_time_info = np.array(
            adata.uns[f"{key_word_0}_{source}_{fate_key}"]["sorted_time_info"]
        )

        row = len(prob_array)
        col = 1
        fig = plt.figure(figsize=(fig_width * col, fig_height * row))
        for k, t_0 in enumerate(sorted_time_info):
            ax = plt.subplot(row, col, 1 + k)
            pl_util.customized_embedding(
                x_emb,
                y_emb,
                prob_array[k],
                ax=ax,
                point_size=point_size,
                title=f"Iter. {k};  t={t_0}",
            )

        plt.tight_layout()
        data_des = adata.uns["data_des"][-1]
        if figure_index != "":
            figure_index == f"_{figure_index}"

        plt.savefig(
            os.path.join(
                f"{settings.figure_path}",
                f"{data_des}_{key_word}_{fate_key}{figure_index}.{settings.file_format_figs}",
            )
        )


def single_cell_transition(
    adata,
    selected_state_id_list,
    source="transition_map",
    map_backward=True,
    savefig=False,
    initial_point_size=3,
    color_bar=True,
):
    """
    Plot transition probability from given initial cell states.

    If `map_backward=True`, plot the probability :math:`T_{ij}` over initial states :math:`i`
    at a given later state :math:`j`. Otherwise, plot the probability :math:`T_{ij}`
    over later states :math:`j` at a fixed initial state :math:`i`.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    selected_state_id_list: `list`
        List of cell id's. Like [0,1,2].
    source: `str`
        The transition map to be used for plotting: {'transition_map',
        'intraclone_transition_map',...}. The actual available
        map depends on adata itself, which can be accessed at adata.uns['available_map']
    map_backward: `bool`, optional (default: True)
        If `map_backward=True`, plot the probability of source states where the current cell state comes from;
        otherwise, plot future state probability starting from given initial state.
    initial_point_size: `int`, optional (default: 3)
        Relative size of the data point for the selected cells.
    save_fig: `bool`, optional (default: False)
        If true, save figure to defined directory at settings.figure_path
    color_bar: `bool`, optional (default: True)
        Plot the color bar.
    """

    hf.check_available_map(adata)
    fig_width = settings.fig_width
    fig_height = settings.fig_height
    point_size = settings.fig_point_size
    if color_bar:
        fig_width = fig_width + 0.5

    if source not in adata.uns["available_map"]:
        logg.error(f"{source} should be among {adata.uns['available_map']}")

    else:
        state_annote = adata.obs["state_info"]
        x_emb = adata.obsm["X_emb"][:, 0]
        y_emb = adata.obsm["X_emb"][:, 1]
        data_des = adata.uns["data_des"][-1]
        figure_path = settings.figure_path

        if not map_backward:
            cell_id_t1 = adata.uns["Tmap_cell_id_t1"]
            cell_id_t2 = adata.uns["Tmap_cell_id_t2"]
            Tmap = adata.uns[source]

        else:
            cell_id_t2 = adata.uns["Tmap_cell_id_t1"]
            cell_id_t1 = adata.uns["Tmap_cell_id_t2"]
            Tmap = adata.uns[source].T

        selected_state_id_list = np.array(selected_state_id_list)
        full_id_list = np.arange(len(cell_id_t1))
        valid_idx = np.in1d(full_id_list, selected_state_id_list)
        if np.sum(valid_idx) < len(selected_state_id_list):
            logg.error(f"Valid id is a integer, ranged in (0,{len(cell_id_t1)-1}).")
            selected_state_id_list = full_id_list[valid_idx]

        if len(selected_state_id_list) == 0:
            logg.error(f"Valid id is a integer, ranged in (0,{len(cell_id_t1)-1}).")
        else:
            if ssp.issparse(Tmap):
                Tmap = Tmap.A

            row = len(selected_state_id_list)
            col = 1
            fig = plt.figure(figsize=(fig_width * col, fig_height * row))

            for j, target_cell_ID in enumerate(selected_state_id_list):
                ax0 = plt.subplot(row, col, col * j + 1)

                if target_cell_ID < Tmap.shape[0]:
                    prob_vec = np.zeros(len(x_emb))
                    prob_vec[cell_id_t2] = Tmap[target_cell_ID, :]
                    prob_vec = prob_vec / np.max(prob_vec)
                    pl_util.customized_embedding(
                        x_emb,
                        y_emb,
                        prob_vec,
                        point_size=point_size,
                        ax=ax0,
                        color_bar=color_bar,
                        color_bar_label="Probability",
                    )

                    ax0.plot(
                        x_emb[cell_id_t1][target_cell_ID],
                        y_emb[cell_id_t1][target_cell_ID],
                        "*b",
                        markersize=initial_point_size * point_size,
                    )

                    if map_backward:
                        ax0.set_title(f"ID (t2): {target_cell_ID}")
                    else:
                        ax0.set_title(f"ID (t1): {target_cell_ID}")

            # if color_bar:
            #     Clb=fig.colorbar(plt.cm.ScalarMappable(cmap=color_map_reds), ax=ax0,label='Probability')

            plt.tight_layout()
            if savefig:
                fig.savefig(
                    os.path.join(
                        figure_path,
                        f"{data_des}_single_cell_transition_{source}_{map_backward}.{settings.file_format_figs}",
                    )
                )



def Clone_path(
    sample_name,
    data_name,
    current_folder_input,
    combined_data_monocle_name,
    time_compute_data,
    ):
    import os
    import anndata
    current_folder = current_folder_input
    if 1:
        sample_name = sample_name
        print("Sample: ",sample_name)
        current_dir = os.getcwd()
        print("Changing the working directory from:", current_dir)
        #eng.run()
        loading_directory = os.path.join(current_folder, sample_name)
        kl_data_menu = "data"
        kl_result_menu = "result"
        data_directory = os.path.join(loading_directory, kl_data_menu)
        result_directory = os.path.join(loading_directory, kl_result_menu)
    print() 
    print("---------------------------------------------------------------------------------------")
    print("Please put the data at: ",loading_directory,"\data")
    print("---------------------------------------------------------------------------------------")
    print()
    print("Current working directory is:", data_directory)
    # 构建data.h5ad文件的完整路径
    h5ad_file_path = os.path.join(data_directory, data_name)
    combined_data_monocle_path = os.path.join(result_directory, combined_data_monocle_name)
    print(h5ad_file_path)
    print(combined_data_monocle_path)
    file_path = h5ad_file_path
    csv_path = combined_data_monocle_path
    h5ad_file_path = os.path.join(data_directory, data_name)
    T_data = os.path.join(data_directory, time_compute_data)
    pseudotime_map = os.path.join(result_directory, 'pseudotime_map.csv')
    

    
    
    return file_path,csv_path,result_directory,T_data,pseudotime_map

'''
    file_path, csv_path = Clone_data(
    sample_name,
    data_name,
    current_folder_input,
    combined_data_monocle_name
    )
'''

    
#Singlecell_barcode( )
def Singlecell_barcode__all(X_clone, T_data, X1, X2, node_size, save_path, 
    file_path, csv_path, 
    with_labels=False, draw_edges=True,
    specified_tree_index=None, 
    specified_node_name=None):
        
    #X_clone是那个稀疏矩阵，T是存伪时间的数组，X1，X2是从emb里提取出的坐标,node_size是画出来的点的直径
    #save_path是要将画出来的每个细胞所属的cluster图的存储路径，with_labels是否加细胞编号, draw_edges是否画箭头
    
    # -*- coding: UTF-8 -*-
    import os
    from re import X
    from xml.dom.expatbuilder import InternalSubsetExtractor
    import random
    import anndata
    import numpy as np
    import scanpy as sc
    import pandas as pd
    import networkx as nx
    import matplotlib.pyplot as plt
    import matplotlib.colors as mcolors
    from scipy.sparse import find
    from scipy.sparse import csr_matrix



    #######################################################################################################
    
    # 路径设置
    #file_path = 'Hematopoiesis_progenitor.h5ad'
    #csv_path = 'G:/ASUMDownloads/combined_data_monocle3.csv'
        
    ####################################################################################################

    
        # 染色方案
    cmap = mcolors.LinearSegmentedColormap.from_list("custom_gray_blue", ["#17324b", "#61ade8"])
    
    index_map = {name: idx for idx, name in enumerate(T_data['cell_name'])}
    rev_index_map = {idx: name for idx, name in enumerate(T_data['cell_name'])}
    # 创建伪时间数组
    T = csv_data['pseudotime']
    lenT = len(T)
    # 创建有向图
    G = nx.DiGraph()

    # 从稀疏矩阵中找出非零元素的位置
    rows, cols, _ = find(X_clone)  

    # 将稀疏矩阵逐点提取，以细胞索引为点，逐个添加有向边
    for row, col in zip(rows, cols):
        if row < lenT and col < lenT:
            if T[row] < T[col]:
                G.add_edge(row, col)
            else:
                G.add_edge(col, row)

    # 识别所有弱连通分量
    weak_components = list(nx.weakly_connected_components(G))
    forest = []

    # 遍历每个弱连通分量
    for component in weak_components:
        subgraph = G.subgraph(component).copy()
        roots = [node for node in subgraph if subgraph.in_degree(node) == 0]
        if not roots:
            roots = [next(iter(component))]

        for root in roots:
            tree = nx.dfs_tree(subgraph, source=root)
            forest.append(tree)

    # 确保保存路径存在
    os.makedirs(save_path, exist_ok=True)

    # 根据指定的参数绘制图像

    trees_to_plot = []
    original_indices = []

    if specified_tree_index is not None:
        if isinstance(specified_tree_index, int):
            specified_tree_index = [specified_tree_index]
        trees_to_plot = [forest[i] for i in specified_tree_index if i < len(forest)]
        original_indices = [i for i in specified_tree_index if i < len(forest)]
    elif specified_node_name is not None:
        if isinstance(specified_node_name, int):
            specified_node_name = [specified_node_name]
        node_indices = [index_map.get(name) for name in specified_node_name if name in index_map]
        for index, tree in enumerate(forest):
            if any(node in tree.nodes for node in node_indices):
                trees_to_plot.append(tree)
                original_indices.append(index)
    else:
        trees_to_plot = forest
        original_indices = list(range(len(forest)))

    all_trees = len(trees_to_plot)
    # 绘制指定的树
    for i, (index, tree) in enumerate(zip(original_indices, trees_to_plot), start=1):
        print(f"Plotting... ( {i} / {all_trees} )", end='\r')  # 输出当前进度
        plt.figure(figsize=(16, 9))
        plt.scatter(X1, X2, c='lightgray', s=node_size, alpha=0.6)
        pos = {node: (X1[node], X2[node]) for node in tree.nodes() if node < lenT}
        node_colors = [T[node] for node in tree.nodes() if node < lenT]
        labels = {node: rev_index_map[node] for node in tree.nodes() if node in rev_index_map}
        if draw_edges:
            nx.draw(tree.subgraph(pos.keys()), pos, node_color=node_colors, cmap=cmap,
                   labels=labels if with_labels else None, edge_color='gray', node_size=node_size, font_size=8)
        else:
            nx.draw_networkx_nodes(tree.subgraph(pos.keys()), pos, node_color=node_colors, cmap=cmap, node_size=node_size)

        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=min(node_colors), vmax=max(node_colors)))
        plt.colorbar(sm, orientation='vertical', label='Pseudotime')
        plt.title(f"UMAP Plot with Pseudotime Coloring - Tree {index}")
        plt.xlabel('UMAP 1')
        plt.ylabel('UMAP 2')
        plt.savefig(os.path.join(save_path, f'Tree_{index}.png'))
        plt.close()


        
        
def Cluster_barcode(
                X_clone, C_data, T_data, 
                  X1, X2, node_size, save_path, 
                  Color="#618a03",
                 specified_cluster_index=None,
                 specified_node_name=None
                 ):

    plt.figure(figsize=(16, 9))
    # 绘制背景点
    plt.scatter(X1, X2, c='lightgray', s=node_size, alpha=0.6)

    # 处理输入参数，允许单个或多个输入
    if isinstance(specified_cluster_index, int):
        specified_cluster_index = [specified_cluster_index]
    if isinstance(specified_node_name, int):
        specified_node_name = [specified_node_name]

    # 创建集合和索引映射
    clusters = {col: set(C_data[col].dropna()) for col in C_data.columns}
    index_map = {name: idx for idx, name in enumerate(T_data['cell_name'])}

    # 处理颜色输入，允许单个颜色或多颜色输入
    if isinstance(Color, str):
        Color = [Color]

    # 确定需要绘制的簇
    if specified_node_name:
        specified_cluster_index = [i for i, cluster in enumerate(clusters.values())
                                   if any(node in cluster for node in specified_node_name)]
    if specified_cluster_index is None:
        specified_cluster_index = range(len(clusters))  # 如果未指定，选择所有簇

    # 绘制指定的簇
    for i, (cluster_name, cluster_set) in enumerate(clusters.items()):
        if i in specified_cluster_index:
            # 获取此簇的节点索引
            node_indices = [index_map[node] for node in cluster_set if node in index_map]
            # 选择颜色，如果颜色不够则随机生成新颜色
            color = Color[i % len(Color)] if i < len(Color) else f"#{random.randint(0, 0xFFFFFF):06x}"
            # 绘制簇
            plt.scatter(X1[node_indices], X2[node_indices], c=color, s=node_size, label=f'Cluster {cluster_name}')
    
    # 添加图例
    plt.legend(title='Clusters')
    plt.title("UMAP Plot with Cluster Coloring")
    plt.xlabel('UMAP 1')
    plt.ylabel('UMAP 2')
    plt.savefig(f'{save_path}/clusters_plot.png')
    plt.close()



    
    
    
    
    
def draw_embedding(
    sample_name,
    data_name,
    current_folder_input,
    combined_data_monocle_name
    
    ):
    
    print()
    
    
    
    
    
def get_row_data(sparse_matrix, row_index):
    if row_index < sparse_matrix.shape[0]:  # 确保行索引有效
        row_start = sparse_matrix.indptr[row_index]
        row_end = sparse_matrix.indptr[row_index + 1]
        columns = sparse_matrix.indices[row_start:row_end]
        values = sparse_matrix.data[row_start:row_end]
        return columns, values
    else:
        return None, None


'''
# 路径设置
file_path = 'Hematopoiesis_progenitor.h5ad'
csv_path = 'G:/ASUMDownloads/combined_data_monocle3.csv'

# 读取H5AD文件
adata = anndata.read_h5ad(file_path)
# 读取CSV
csv_data = pd.read_csv(csv_path)
# 读取 CSV 文件
cluster_data = pd.read_csv('G:/ASUMDownloads/var3_intervals.csv')

# 获取X_clone矩阵
X_clone = adata.obsm['X_clone']

#读取umap坐标 
umap_df = pd.DataFrame(adata.obsm['X_emb'], columns=['X1', 'X2'])
X1 = umap_df['X1']
X2 = umap_df['X2']

node_size=4

save_path='G:/ASUMDownloads/dataplot/'
indexA = 2097
indexB = 392
# 获取第2097行和第392行的数据
columns_A, values_A = get_row_data(X_clone, indexA)
columns_B, values_B = get_row_data(X_clone, indexB)

# 输出结果
print(f"Columns and values in row {indexA}:")
print("Columns:", columns_A)
print("Values:", values_A)

print(f"Columns and values in row {indexB}:")
print("Columns:", columns_B)
print("Values:", values_B)

# 示例调用
plot_Clusters(X_clone,cluster_data, csv_data, X1, X2, node_size, save_path, Color="#61ade8",specified_node_name=37803)
# 稀疏矩阵 伪时间 X.Y的UMAP,节点大小,存储路径,是否显示标签,是否显示边
#plot_Colormap(X_clone,csv_data,X1,X2,node_size,save_path,with_labels=True,draw_edges=True,specified_tree_index=232)
#plot_Colormap(X_clone,csv_data,X1,X2,node_size,save_path,with_labels=True,draw_edges=True,specified_tree_index=[233,234])
#plot_Colormap(X_clone,csv_data,X1,X2,node_size,save_path,with_labels=False,draw_edges=False,specified_node_name=32689)
#plot_Colormap(X_clone,csv_data,X1,X2,node_size,save_path,with_labels=True,draw_edges=True,specified_node_name=[5314,4465])
#plt.ion()
'''


# -*- coding: utf-8 -*-

import os
import time
from pathlib import Path, PurePath

import numpy as np
import pandas as pd
import scanpy as sc
import scipy.sparse as ssp
from matplotlib import pyplot as plt

from .. import help_functions as hf
from .. import logging as logg
from .. import plotting as pl
from .. import settings


def initialize_adata_object(
    adata=None,
    X_state=None,
    cell_names=None,
    gene_names=None,
    time_info=None,
    X_clone=None,
    X_pca=None,
    X_emb=None,
    state_info=None,
    data_des="cospar",
):
    """
    Initialized the :class:`~anndata.AnnData` object.

    The principal components (`X_pca`), 2-d embedding (`X_emb`), and
    `state_info` can be provided upfront, or generated in the next step.
    The clonal information (`X_clone`) is also optional.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    X_state: `np.ndarray` or `sp.spmatrix`, optional (default: None)
        The count matrix for state information.
        Rows correspond to cells and columns to genes. If adata is provided, this
        is not necessary.
    cell_names: `np.ndarray`, optional (default: None)
        An array of cell ids. If adata is provided, this
        is not necessary.
    gene_names: `np.ndarray`, optional (default: None)
        An array of gene names. If adata is provided, this
        is not necessary.
    time_info: `np.ndarray`
        Time annotation for each cell in `str`,like 'Day27' or 'D27'.
        However, it can also contain other sample_info,
        like 'GFP+_day27', and 'GFP-_day27'.
    X_clone: `sp.spmatrix` or `np.ndarray`, optional (default: None)
        The clonal data matrix, with the row in cell id, and column in barcode id.
        For evolvable barcoding, a cell may carry several different barcode id.
        Currently, we assume each entry is 0 or 1.
    X_pca: `np.ndarray`, optional (default: None)
        A matrix of the shape n_cell*n_pct. Used for computing similarity matrices.
    X_emb: `np.ndarray`, optional (default: None)
        Two-dimensional matrix for embedding.
    state_info: `np.ndarray`, optional (default: None)
        The classification and annotation for each cell state.
        This will be used only after the map is created. Can be adjusted later.
    data_des: `str`, optional (default:'cospar')
        This is just a name to label/distinguish this data.
        Will be used for saving the results. It should be a unique
        name for a new dataset stored in the same folder to avoid conflicts.

    Returns
    -------
    Generate an :class:`~anndata.AnnData` object with the following entries
    adata.var_names: gene names
    adata.obs_names: cell IDs
    obs: 'time_info', 'state_info'
    uns: 'data_des', 'clonal_time_points', 'time_ordering'
    obsm: 'X_clone', 'X_pca', 'X_emb'
    """

    data_path = settings.data_path
    figure_path = settings.figure_path

    ### making folders
    if not Path(data_path).is_dir():
        logg.info(f"creating directory {data_path} for saving data")
        Path(data_path).mkdir(parents=True)

    if not Path(figure_path).is_dir():
        logg.info(f"creating directory {figure_path} for saving figures")
        Path(figure_path).mkdir(parents=True)

    if adata is None:
        if (X_state is not None) and (gene_names) is not None:
            logg.info("Create new anndata object")
            adata = sc.AnnData(ssp.csr_matrix(X_state))
            adata.var_names = list(gene_names)
        else:
            raise ValueError(
                "If adata is not provided, X_state and gene_names must be provided. Abort initialization."
            )

    if cell_names is not None:
        adata.obs_names = list(cell_names)

    adata.var_names_make_unique()  # this is needed for get_X_pca()
    adata.obs_names_make_unique()
    if time_info is None:
        if "time_info" not in adata.obs.keys():
            logg.warn(
                "time_info not provided. Initialize the time info to be ['0','0','0',...]"
            )
            time_info = ["0" for xx in range(adata.shape[0])]
            adata.obs["time_info"] = pd.Categorical(time_info)

    else:
        time_info = np.array(time_info)
        time_info = time_info.astype(str)
        adata.obs["time_info"] = pd.Categorical(time_info)

    # we do not remove zero-sized clones here as in some case, we want
    # to use this package even if we do not have clonal data.
    # Removing zero-sized clones will be handled downstream when we have to
    if X_clone is None:
        if "X_clone" not in adata.obsm.keys():
            X_clone = np.zeros((adata.shape[0], 1))
        else:
            X_clone = adata.obsm["X_clone"]
    else:
        if X_clone.shape[0] != adata.shape[0]:
            raise ValueError(
                "X_clone.shape[0] not equal to cell number. Abort initialization."
            )

    # remove clones without a cell
    if X_clone.shape[1] > 1:
        ini_clone_N = X_clone.shape[1]
        X_clone = ssp.csr_matrix(X_clone)
        valid_clone_id = np.nonzero(X_clone.sum(0).A.flatten() > 0)[0]
        X_clone_temp = X_clone[:, valid_clone_id]
        adata.obsm["X_clone"] = ssp.csr_matrix(X_clone_temp)
        if adata.obsm["X_clone"].shape[1] < ini_clone_N:
            logg.info("Clones without any cells are removed.")
    else:
        adata.obsm["X_clone"] = ssp.csr_matrix(X_clone)

    if "data_des" not in adata.uns.keys():
        adata.uns["data_des"] = [data_des]

    if X_pca is not None:
        if X_pca.shape[0] == adata.shape[0]:
            adata.obsm["X_pca"] = np.array(X_pca)
        else:
            logg.error("X_pca.shape[0] not equal to cell number")
    else:
        if "X_pca" not in adata.obsm.keys():
            logg.warn(
                "X_pca not provided. Downstream processing is needed to generate X_pca before computing the transition map."
            )

    if state_info is not None:
        if len(state_info) == adata.shape[0]:
            adata.obs["state_info"] = pd.Categorical(state_info)
        else:
            logg.error("state_info length not equal to cell number")
    else:
        if "state_info" not in adata.obs.keys():
            if "leiden" in adata.obs.keys():
                logg.warn(
                    "Use adata.obs['leiden'] as the default state annotation. If not desirable, please use downstream analysis to improve the annotation."
                )
                adata.obs["state_info"] = adata.obs["leiden"]
            else:
                logg.warn(
                    "state_info not provided. Downstream processing is needed before analyzing the transition map."
                )

    if X_emb is not None:
        if X_emb.shape[0] == adata.shape[0]:
            adata.obsm["X_emb"] = X_emb
        else:
            logg.error("X_emb.shape[0] not equal to cell number")
    else:
        if "X_emb" not in adata.obsm.keys():
            if "X_umap" in adata.obsm.keys():
                logg.warn("Use X_umap as the default embedding")
                adata.obsm["X_emb"] = adata.obsm["X_umap"]
            else:
                logg.warn(
                    "X_emb not provided. Downstream processing is needed before analyzing the transition map."
                )

    # logg.info(f"All time points: {set(adata.obs['time_info'])}")
    # logg.info(f"Time points with clonal info: {set(adata.uns['clonal_time_points'])}")

    time_ordering = np.sort(list(set(adata.obs["time_info"])))
    adata.uns["time_ordering"] = time_ordering
    hf.check_available_clonal_info(adata)
    logg.info(f"Time points with clonal info: {adata.uns['clonal_time_points']}")
    logg.warn(
        f"Default ascending order of time points are: {adata.uns['time_ordering']}. If not correct, run cs.hf.update_time_ordering for correction."
    )

    logg.warn(f"Please make sure that the count matrix adata.X is NOT log-transformed.")

    return adata


def get_highly_variable_genes(
    adata,
    normalized_counts_per_cell=10000,
    min_counts=3,
    min_cells=3,
    min_gene_vscore_pctl=85,
):
    """
    Get highly variable genes.

    We assume that data preprocessing are already done, like removing low quality cells.
    It first perform count normalization, then variable gene selection.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    normalized_counts_per_cell: `int`, optional (default: 1000)
        count matrix normalization
    min_counts: `int`, optional (default: 3)
        Minimum number of UMIs per cell to be considered for selecting highly variable genes.
    min_cells: `int`, optional (default: 3)
        Minimum number of cells per gene to be considered for selecting highly variable genes.
    min_gene_vscore_pctl: `int`, optional (default: 85)
        Gene expression variability threshold, in the unit of percentile,
        for selecting highly variable genes. Range: [0,100], with a higher
        number selecting more variable genes.

    Returns
    -------
    Modify adata.var['highly_variable'].
    If 'highly_variable' existed before, save a copy at adata.obs['highly_variable_old']
    """

    sc.pp.normalize_per_cell(adata, counts_per_cell_after=normalized_counts_per_cell)

    verbose = logg._settings_verbosity_greater_or_equal_than(
        2
    )  # the highest level is 3

    logg.info("Finding highly variable genes...")
    gene_list = adata.var_names
    gene_idx = hf.filter_genes(
        adata.X,
        min_counts=min_counts,
        min_cells=min_cells,
        min_vscore_pctl=min_gene_vscore_pctl,
        show_vscore_plot=verbose,
    )
    highvar_genes = gene_list[gene_idx]

    if "highly_variable" in adata.var.keys():
        adata.var["highly_variable_old"] = adata.var["highly_variable"].copy()

    adata.var["highly_variable"] = False
    adata.var.loc[highvar_genes, "highly_variable"] = True
    logg.info(f"Keeping {len(highvar_genes)} genes")


def remove_cell_cycle_correlated_genes(
    adata,
    cycling_gene_list=[
        "Ube2c",
        "Hmgb2",
        "Hmgn2",
        "Tuba1b",
        "Ccnb1",
        "Tubb5",
        "Top2a",
        "Tubb4b",
    ],
    corr_threshold=0.1,
    confirm_change=False,
):
    """
    Remove cell-cycle correlated genes.

    Take pre-selected highly variable genes, and compute their correlation with
    the set of cell cycle genes. If confirm_change=True, remove those having absolute correlation
    score are above given correlation threshold. It is a prerequisite to run
    :func:`get_highly_variable_genes` first.

    Warning: the default cell cycle gene sets are from mouse genes. Please convert them
    to upper case if you want to apply it to human data. Also, consider using your own
    gene sets.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    cycling_gene_list: `list`, optional
        A list of cell cycle correlated genes to compute correlation with.
    corr_threshold: `float`, optional (default: 0.1)
        Highly variable genes with absolute correlation score about this
        threshold will be removed from the highly variable gene list.
    confirm_change: `bool`, optional (default: False)
        If set True, adata.var['highly_variable'] will be updated to exclude
        cell cycle correlated genes.
    """

    if "highly_variable" not in adata.var.keys():
        logg.error(
            "Did not find highly variable genes index in adata.var['highly_variable']\n"
            "Please run cs.pl.get_highly_variable_genes first!"
        )

    else:
        gene_list = np.array(adata.var_names)

        cycling_gene_idx = np.in1d(gene_list, cycling_gene_list)
        if np.sum(cycling_gene_idx) != len(cycling_gene_list):
            logg.error(
                f"Provided cyclcing genes: {cycling_gene_list}\n"
                f"They are for mouse genes. Only {np.sum(cycling_gene_idx)} found in the reference gene list."
            )
        else:
            E = adata.X
            cycling_expression = E[:, cycling_gene_idx].A.T

            highvar_genes_idx = np.array(adata.var["highly_variable"])
            highvar_genes = gene_list[highvar_genes_idx]
            test_expression = E[:, highvar_genes_idx].A.T

            cell_cycle_corr = hf.corr2_coeff(test_expression, cycling_expression)

            # adata.uns['cycling_correlation']

            if not confirm_change:
                max_corr = np.max(abs(cell_cycle_corr), 1)
                fig = plt.figure(figsize=(4, 3.5))
                ax0 = plt.subplot(1, 1, 1)
                ax0.hist(max_corr, 100)
                ax0.set_xlabel("Max. corr. with cycling genes")
                ax0.set_ylabel("Histogram")

                logg.info(
                    "adata.var['highly_variable'] not updated.\n"
                    "Please choose corr_threshold properly, and set confirm_change=True"
                )
            else:
                max_corr = np.max(abs(cell_cycle_corr), 1)
                noCycle_idx = max_corr < corr_threshold

                highvar_genes_noCycle = highvar_genes[noCycle_idx]
                logg.info(
                    f"Number of selected non-cycling highly variable genes: {len(highvar_genes_noCycle)}\n"
                    f"Remove {np.sum(~noCycle_idx)} cell cycle correlated genes."
                )

                highvar_genes_noCycle_idx = np.in1d(gene_list, highvar_genes_noCycle)
                adata.var["highly_variable"] = highvar_genes_noCycle_idx
                logg.info("adata.var['highly_variable'] updated")


def get_X_pca(adata, n_pca_comp=40):
    """
    Get X_pca.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    n_pca_comp: int, optional (default: 40)
        Number of top principle components to keep

    Returns
    -------
    Modify adata.obsm['X_pca'].
    If 'X_pca' existed before, save a copy at  adata.obs['X_pca_old']
    """

    logg.warn(
        f"get_X_pca assumes that the count matrix adata.X is NOT log-transformed."
    )
    if "highly_variable" not in adata.var.keys():
        if adata.shape[1] > 100:
            raise ValueError(
                "Did not find highly variable genes index in adata.var['highly_variable']\n"
                "Please run cs.pl.get_highly_variable_genes first!"
            )

        else:
            logg.warn(
                "Did not find highly variable genes index in adata.var['highly_variable']\n"
                "Compute X_pca with all genes"
            )

            highvar_genes_idx = np.ones(adata.shape[1]).astype(bool)
    else:
        highvar_genes_idx = np.array(adata.var["highly_variable"])

    if n_pca_comp > adata.shape[1]:
        n_pca_comp = adata.shape[1]

    if "X_pca" in adata.obsm.keys():
        adata.obsm["X_pca_old"] = adata.obsm["X_pca"].copy()

    gene_list = np.array(adata.var_names)
    highvar_genes = gene_list[highvar_genes_idx]

    zero_idx = adata[:, highvar_genes].X.sum(0).A.flatten() == 0
    if np.sum(zero_idx) > 0:
        logg.warn(
            f"Genes {highvar_genes[zero_idx]} are not expressed. They are ignored."
        )
        highvar_genes = highvar_genes[~zero_idx]

    adata.obsm["X_pca"] = hf.get_pca(
        adata[:, highvar_genes].X,
        numpc=n_pca_comp,
        keep_sparse=False,
        normalize=True,
        random_state=0,
    )


def get_X_emb(adata, n_neighbors=20, umap_min_dist=0.3):
    """
    Get X_emb using :func:`scanpy.tl.umap`

    We assume that X_pca is computed. It first runs KNN graph construction.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    n_neighbors: `int`, optional (default: 20)
        Neighber number for constructing the KNN graph, using the UMAP method.
    umap_min_dist: `float`, optional (default: 0.3)
        The effective minimum distance between embedded points.

    Returns
    -------
    Modify adata.obsm['X_emb'].
    If 'X_emb' existed before, save a copy at  adata.obs['X_emb_old']
    """

    if not ("X_pca" in adata.obsm.keys()):
        logg.error("*X_pca* missing from adata.obsm... abort the operation")
    else:
        # Number of neighbors for KNN graph construction
        sc.pp.neighbors(adata, n_neighbors=n_neighbors)
        sc.tl.umap(adata, min_dist=umap_min_dist)
        if "X_emb" in adata.obsm.keys():
            adata.obsm["X_emb_old"] = adata.obsm["X_emb"].copy()

        adata.obsm["X_emb"] = adata.obsm["X_umap"]


def get_state_info(adata, n_neighbors=20, resolution=0.5):
    """
    Update `state_info` using :func:`scanpy.tl.leiden`

    We assume that `adata.obsm['X_pca']` exists. It first runs KNN graph construction.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    n_neighbors: `int`, optional (default: 20)
        Neighber number for constructing the KNN graph, using the UMAP method.
    resolution: `float`, optional (default: 0.5)
        A parameter value controlling the coarseness of the clustering.
        Higher values lead to more clusters.

    Returns
    -------
    Modify adata.obs['state_info'].
    If 'state_info' existed before, save a copy at  adata.obs['state_info_old']
    """
    sc.pp.neighbors(adata, n_neighbors=n_neighbors)
    if not ("X_pca" in adata.obsm.keys()):
        logg.error("*X_pca* missing from adata.obsm... abort the operation")
    else:
        # Number of neighbors for KNN graph construction
        sc.tl.leiden(adata, resolution=resolution)
        if "state_info" in adata.obs.keys():
            adata.obs["state_info_old"] = adata.obs["state_info"].copy()

        adata.obs["state_info"] = adata.obs["leiden"]


def get_X_clone(
    adata,
    clone_data_cell_id,
    clone_data_barcode_id,
    reference_cell_id=None,
    reference_clone_id=None,
):
    """
    Build the X_clone matrix from data.

    Convert the raw clonal data table (long format): [clone_data_cell_id,clone_data_barcode_id]
    to X_clone (wide format) based on the unique cell id ordering in adata

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        We assume that adata.obs_names have been initialized.
    clone_data_cell_id: `list`
        The list of cell id for each corresponding sequenced barcode.
    clone_data_barcode_id: `list`
        The list of barcode id from sequencing. It has the same shape as clone_data_cell_id.
    reference_cell_id: `list`, optional (default: None)
        A list of uniuqe cell id. X_clone will be generated based on its cell id ordering.
        This has to be provided to match the cell ordering in the adata object.
    reference_clone_id: `list`, optional (default: None)
        A list of uniuqe clone id. If provided, X_clone will be generated based on its barcode ordering.

    Returns
    -------
    Update adata.obsm['X_clone']: `sp.spmatrix`
    """

    if reference_cell_id is None:
        reference_cell_id = np.array(adata.obs_names)
    else:
        if len(reference_cell_id) == adata.shape[0]:
            adata.obs_names = reference_cell_id
        else:
            logg.error("reference_cell_id does not have the size of adata.shape[0]")

    (
        X_clone,
        reference_clone_id_1,
        reference_cell_id,
    ) = hf.get_X_clone_with_reference_ordering(
        clone_data_cell_id,
        clone_data_barcode_id,
        reference_cell_id=reference_cell_id,
        reference_clone_id=reference_clone_id,
    )
    if X_clone.sum() == 0:
        raise ValueError(
            "X_clone is not initialized. Zero entries detected.\n"
            "Possible reason: the reference_cell_id (from adata.obs_names or directly provided) is not right"
        )
    else:
        if reference_clone_id is None:
            logg.info('reorder clones')
            reference_clone_id_1=np.array(reference_clone_id_1)
            orig_order=reference_clone_id_1
            new_order=sorted(reference_clone_id_1)
            reorder_index=[]
            for i, x in enumerate(new_order):
                j0=np.nonzero(orig_order==x)[0][0]
                reorder_index.append(j0)

            X_clone=X_clone[:,reorder_index]
            reference_clone_id_1=reference_clone_id_1[reorder_index]
        adata.obsm["X_clone"] = X_clone
        adata.uns["clone_id"] = reference_clone_id_1

        # hf.check_available_clonal_info(adata)


############# refine clusters for state_info


def refine_state_info_by_leiden_clustering(
    adata,
    selected_key="state_info",
    selected_values=None,
    resolution=0.5,
    n_neighbors=20,
    confirm_change=False,
    cluster_name_prefix="S",
):
    """
    Refine state info by clustering states at given time points.

    Select states at desired time points to improve the clustering. When
    first run, set confirm_change=False. Only when you are happy with the
    result, set confirm_change=True to update adata.obs['state_info'].
    The original state_info will be stored at adata.obs['old_state_info'].

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    selected_key:
        A key in adata.obs, including 'state_info', or 'time_info'
    selected_values: `list`, optional (default: include all)
        A list of clusters/time_points for further sub-clustering. Should be
        among adata.obs[selected_key].
    adata: :class:`~anndata.AnnData` object
    n_neighbors: `int`, optional (default: 20)
        Neighber number for constructing the KNN graph, using the UMAP method.
    resolution: `float`, optional (default: 0.5)
        A parameter value controlling the coarseness of the clustering.
        Higher values lead to more clusters.
    confirm_change: `bool`, optional (default: False)
        If True, update adata.obs['state_info']
    cluster_name_prefix: `str`, optional (default: 'S')
        prefix for the new cluster name to distinguish it from
        existing cluster names.

    Returns
    -------
    Update adata.obs['state_info'] if confirm_change=True.
    """

    time_info = adata.obs[selected_key]
    available_time_points = list(set(time_info))

    if selected_values == None:
        selected_values = available_time_points
    if type(selected_values) == str:
        selected_values = [selected_values]

    if np.sum(np.in1d(selected_values, available_time_points)) != len(selected_values):
        logg.error(
            f"Selected time points not available. Please select from {available_time_points}"
        )

    else:
        sp_idx = np.zeros(adata.shape[0], dtype=bool)
        for xx in selected_values:
            idx = time_info == xx
            sp_idx[idx] = True

        adata_sp = adata[sp_idx]
        sc.pp.neighbors(adata_sp, n_neighbors=n_neighbors)
        sc.tl.leiden(adata_sp, resolution=resolution)

        pl.embedding(adata_sp, color="leiden")

        if confirm_change:
            logg.info("Change state annotation at adata.obs['state_info']")

            if "state_info" in adata.obs.keys():
                adata.obs["state_info_old"] = adata.obs["state_info"].copy()

            orig_state_annot = np.array(adata.obs["state_info"])
            temp_array = np.array(adata_sp.obs["leiden"])
            for j in range(len(temp_array)):
                temp_array[j] = cluster_name_prefix + temp_array[j]

            orig_state_annot[sp_idx] = temp_array
            adata.obs["state_info"] = pd.Categorical(orig_state_annot)
            pl.embedding(adata, color="state_info")


def refine_state_info_by_marker_genes(
    adata,
    marker_genes,
    express_threshold=0.1,
    selected_key="state_info",
    selected_values=None,
    new_cluster_name="new_cluster",
    confirm_change=False,
    add_neighbor_N=5,
):
    """
    Refine state info according to marker gene expression.

    In this method, a state is selected if it expresses all genes in the list
    of marker_genes, and the expression are above the relative `express_threshold`.
    You can also specify which time point you want to focus on. In addition, we also include
    cell states neighboring to these valid states to smooth the selection
    (controlled by add_neighbor_N).

    When you run it the first time, set confirm_change=False. Only when you are happy with
    the result, set confirm_change=True to update the adata.obs['state_info'].
    The original state_info will be stored at adata.obs['old_state_info'].

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    marker_genes: `list` or 'str'
        List of marker genes to be used for defining cell states.
    express_threshold: `float`, optional (default: 0.1)
        Relative threshold of marker gene expression, in the range [0,1].
        A state must have an expression above this threshold for all genes
        to be included.
    selected_key:
        A key in adata.obs, including 'state_info', or 'time_info'
    selected_values: `list`, optional (default: include all)
        A list of clusters/time_points for further sub-clustering. Should be
        among adata.obs[selected_key].
    new_cluster_name: `str`, optional (default: 'new_cluster')
    confirm_change: `bool`, optional (default: False)
        If True, update adata.obs['state_info'].
    add_neighbor_N: `int`, optional (default: 5)
        Add to the new cluster neighboring cells of a qualified
        high-expressing state according to the KNN graph
        with K=add_neighbor_N.

    Returns
    -------
    Update the adata.obs['state_info'] if confirm_change=True.
    """

    time_info = adata.obs[selected_key]
    x_emb = adata.obsm["X_emb"][:, 0]
    y_emb = adata.obsm["X_emb"][:, 1]
    available_time_points = list(set(time_info))

    if selected_values == None:
        selected_values = available_time_points
    if type(selected_values) == str:
        selected_values = [selected_values]

    sp_idx = np.zeros(adata.shape[0], dtype=bool)
    for xx in selected_values:
        idx = time_info == xx
        sp_idx[idx] = True

    # add gene constraints
    selected_states_idx = np.ones(adata.shape[0], dtype=bool)
    gene_list = list(adata.var_names)
    tot_name = ""

    if type(marker_genes) == str:
        marker_genes = [marker_genes]

    for marker_gene_temp in marker_genes:
        if marker_gene_temp in gene_list:
            expression = adata.obs_vector(marker_gene_temp)
            thresh = express_threshold * np.max(expression)
            idx = expression > thresh
            selected_states_idx = selected_states_idx & idx

            tot_name = tot_name + marker_gene_temp

    # add temporal constraint
    selected_states_idx[~sp_idx] = 0

    if np.sum(selected_states_idx) > 0:
        # add neighboring cells to smooth selected cells (in case the expression is sparse)
        selected_states_idx = hf.add_neighboring_cells_to_a_map(
            selected_states_idx, adata, neighbor_N=add_neighbor_N
        )

        fig_width = settings.fig_width
        fig_height = settings.fig_height
        fig = plt.figure(figsize=(fig_width, fig_height))
        ax = plt.subplot(1, 1, 1)
        pl.customized_embedding(x_emb, y_emb, selected_states_idx, ax=ax)
        ax.set_title(f"{tot_name}; Cell #: {np.sum(selected_states_idx)}")
        # print(f"Selected cell state number: {np.sum(selected_states_idx)}")

        if confirm_change:
            if "state_info" in adata.obs.keys():
                adata.obs["state_info_old"] = adata.obs["state_info"].copy()

            logg.info("Change state annotation at adata.obs['state_info']")
            if new_cluster_name == "":
                new_cluster_name = marker_genes[0]

            orig_state_annot = np.array(adata.obs["state_info"])
            orig_state_annot[selected_states_idx] = np.array(
                [new_cluster_name for j in range(np.sum(selected_states_idx))]
            )
            adata.obs["state_info"] = pd.Categorical(orig_state_annot)
            pl.embedding(adata, color="state_info")
    else:
        logg.error("Either the gene names or the time point names are not right.")


def filter_clone_size(adata, lower_cutoff=2, upper_cutoff=None):
    """
    Filter out barcodes with cells outside of the range (lower_cutoff, upper_cutoff)
    """
    X_clone_0 = adata.obsm["X_clone"]
    clone_size = X_clone_0.sum(0).A.flatten()
    clone_idx = clone_size >= lower_cutoff
    if upper_cutoff is not None:
        clone_idx = clone_idx & (clone_size <= upper_cutoff)

    adata.obsm["X_clone"] = X_clone_0[:, clone_idx]


def filter_cells_with_many_barcodes(adata, max_barcodes=10):
    """
    Filter out cells with more than `max_barcodes` of barcodes
    """
    X_clone_0 = adata.obsm["X_clone"].A.copy()
    barcode_N = X_clone_0.sum(1)
    cell_idx = barcode_N > max_barcodes
    X_clone_0[cell_idx] = 0
    adata.obsm["X_clone"] = ssp.csr_matrix(X_clone_0)


def filter_nonclonal_cells(adata):
    sp_idx = adata.obsm["X_clone"].sum(1).A.flatten() > 0
    logg.info(f"Original cell number {len(sp_idx)}; New cell number {np.sum(sp_idx)}")
    return adata[sp_idx]



import os
import time
from logging import raiseExceptions

import numpy as np
import pandas as pd
import scipy.sparse as ssp
import scipy.stats as stats
import seaborn as sns
import statsmodels.sandbox.stats.multicomp
from ete3 import Tree
from matplotlib import pyplot as plt
from scipy.cluster import hierarchy

# from plotnine import *
from sklearn.manifold import SpectralEmbedding
from tqdm import tqdm

from cospar.tool import _utils as tl_util

from .. import help_functions as hf
from .. import logging as logg
from .. import plotting as pl
from .. import settings


def clonal_fate_bias(
    adata,
    selected_fate="",
    alternative="two-sided",
    multiple_hypothesis_correction=False,
):
    """
    Compute clonal fate bias towards a cluster.

    The clonal fate bias is -log(Q-value). We calculated a P-value that
    that a clone is enriched (or depleted) in a fate, using Fisher-Exact
    test (accounting for clone size). The P-value is then corrected to
    give a Q-value by Benjamini-Hochberg procedure. The alternative
    hypothesis options are: {'two-sided', 'greater', 'less'}.
    The default is 'two-sided'. With two-sided, we simply take less likely
    scenario either 'greater' and 'less' hypothesis.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    selected_fate: `str`
        The targeted fate cluster, from adata.obs['state_info'].
    alternative: `str`, optional (default: 'two-sided')
        Defines the alternative hypothesis. The following options are
        available (default is ‘two-sided’): ‘two-sided’;
        ‘less’: one-sided; ‘greater’: one-sided

    Returns
    -------
    result: `pd.DataFrame`
    """

    if alternative not in ["two-sided", "less", "greater"]:
        logg.warn(
            "alternative not in ['two-sided','less','greater']. Use 'two-sided' instead."
        )
        alternative = "two-sided"

    state_info = adata.obs["state_info"]
    X_clone = adata.obsm["X_clone"]
    clone_N = X_clone.shape[1]
    cell_N = X_clone.shape[0]

    if type(selected_fate) == list:
        selected_fate = [selected_fate]

    if type(selected_fate) == str:
        selected_fate = [selected_fate]

    (
        mega_cluster_list,
        valid_fate_list,
        fate_array_flat,
        sel_index_list,
    ) = hf.analyze_selected_fates(state_info, selected_fate)
    if len(mega_cluster_list) == 0:
        logg.error("No cells selected. Computation aborted!")
        return None, None
    else:
        fate_name = mega_cluster_list[0]
        target_idx = sel_index_list[0]  # whether or not a cell is in a target cluster

        ## target clone
        target_ratio_array = np.zeros(clone_N)
        P_value = np.zeros(clone_N) + 0.5

        for m in tqdm(range(clone_N)):
            clone_cell_idx = (
                X_clone[:, m].sum(1).A > 0
            ).flatten()  # cell index of clone m
            clone_size = np.sum(clone_cell_idx)  # size of this clone

            if clone_size > 0:
                target_ratio = (
                    np.sum(target_idx[clone_cell_idx]) / clone_size
                )  # fraction of cells from this clone in the target cluster
                target_ratio_array[m] = target_ratio
                cell_N_in_target = np.sum(
                    target_idx[clone_cell_idx]
                )  # number of cells from this clone that are in the target cluster
                target_size = np.sum(target_idx)
                clone_size = np.sum(clone_cell_idx)

                p1 = stats.hypergeom(M=cell_N, n=target_size, N=clone_size).cdf(
                    cell_N_in_target
                )  # observing events <=x
                p2 = stats.hypergeom(M=cell_N, n=target_size, N=clone_size).sf(
                    cell_N_in_target - 1
                )  # observing events >=x
                if alternative == "greater":
                    pvalue = p1
                elif alternative == "less":
                    pvalue = p2
                else:
                    pvalue = np.min(
                        [p1, p2]
                    )  # comparing the two hypothesis, and selecting the less likely one

                P_value[m] = pvalue

        clone_size_array = X_clone.sum(0).A.flatten()
        result = pd.DataFrame(
            {
                "clone_id": np.arange(len(clone_size_array)),
                "clone_size": clone_size_array,
                "P_value": P_value,
                "clonal_fraction_in_target_fate": target_ratio_array,
            }
        )

        if multiple_hypothesis_correction:

            def hypothesis_testing(pv):
                qv = statsmodels.sandbox.stats.multicomp.multipletests(
                    list(pv), alpha=0.05, method="fdr_bh"
                )[1]
                return qv

            df_list = []
            for x in result["clone_size"].unique():
                # display(result[result['P_value']==1])
                df_tmp = result[(result["clone_size"] == x)]
                if len(df_tmp) > 0:
                    df_tmp["Q_value"] = hypothesis_testing(df_tmp["P_value"])
                    df_list.append(df_tmp)
            df_final = pd.concat(df_list)
            df_final["Fate_bias"] = -np.log10(df_final["Q_value"])
        else:
            df_final = result
            df_final["Fate_bias"] = -np.log10(df_final["P_value"])

        df_final = df_final.sort_values("Fate_bias", ascending=False)

        adata.uns["clonal_fate_bias"] = df_final[df_final["clone_size"] > 0]
        logg.info("Data saved at adata.uns['clonal_fate_bias']")


def identify_persistent_clones(adata):
    time_info = np.array(adata.obs["time_info"])
    X_clone = adata.obsm["X_clone"]
    unique_time_info = list(set(time_info))
    persistent_clone_idx = np.ones(X_clone.shape[1]).astype(bool)
    for t in unique_time_info:
        persistent_clone_idx = persistent_clone_idx & (
            X_clone[time_info == t].sum(0).A.flatten() > 0
        )
    persistent_clone_ids = np.nonzero(persistent_clone_idx)[0]
    return persistent_clone_ids


def fate_biased_clones(
    adata,
    selected_fate,
    fraction_threshold=0.1,
    FDR_threshold=0.05,
    persistent_clones=False,
):
    """
    Find clones that significantly biased towards a given fate

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    selected_fate: `str`
        The targeted fate cluster, from adata.obs['state_info'].
    fraction_threshold: float
        The fraction of cells in the target fate cluster for a clone to be included.
    FDR_threshold:
        False discovery rate for identifying the fate biased clone.
    persistent_clones: bool
        True: find clones that are shared across time points.

    Returns
    -------
    valid_clone_ids:
        List of valid clone ids.
    """
    clonal_fate_bias(adata, selected_fate=selected_fate)
    result = adata.uns["clonal_fate_bias"]
    valid_clone_ids = list(
        result[
            (result.Q_value < FDR_threshold)
            & (result.clonal_fraction_in_target_fate > fraction_threshold)
        ]["Clone_ID"]
    )

    if persistent_clones:
        persistent_ids = identify_persistent_clones(adata)
        valid_clone_ids = list(set(valid_clone_ids).intersection(set(persistent_ids)))

    return valid_clone_ids


def get_coarse_grained_X_clone_for_clone_assignment(adata, cluster_key="leiden"):
    """
    This is used in the context of clonal data demultiplexing, like in the multi-seq protocol, where
    we each cell is labeled by one (or several) hash tag, and the barcoding process is noisy. We can look
    at the clusters at the barcode space, and we can assume that, if a cluster is clearly distinct, we
    simply assume that it belongs to a unique clone.
    """
    cell_type_N = []
    selected_fates = list(set(adata.obs[cluster_key]))
    for x in selected_fates:
        temp = np.sum(np.array(adata.obs[cluster_key]) == x)
        cell_type_N.append(temp)

    df_data = pd.DataFrame(adata.obsm["X_clone"].A)
    df_data["cluster"] = list(adata.obs[cluster_key])
    df_X_cluster_0 = df_data.groupby("cluster").sum()

    coarse_X_clone = df_X_cluster_0.to_numpy()

    # normalize cluster wise
    sum_X = 10 ** (-10) + np.array(
        cell_type_N
    )  # we use the cells with/without clonal information
    norm_X_cluster = coarse_X_clone / sum_X[:, np.newaxis]

    # # normalize clone wise
    # sum_X_1 = 10**(-10)+norm_X_cluster.sum(0)
    # norm_X_cluster_clone=norm_X_cluster/sum_X_1[np.newaxis,:]

    # if normalize_per_cluster:
    #     # Finally, normalize cluster wise, as we are interested where each cluster comes from
    #     # The risk here is that if a cluster is uniquely labled by only one clone, but the number
    #     # of clonal cells there is very small, it could lead to errors.
    logg.info("only normlaize per cluster")
    sum_X_2 = 10 ** (-10) + norm_X_cluster.sum(1)
    norm_X_cluster_clone = norm_X_cluster / sum_X_2[:, np.newaxis]

    df_X_cluster = pd.DataFrame(
        norm_X_cluster_clone,
        columns=[f"clone_{j}" for j in range(norm_X_cluster_clone.shape[1])],
    )
    df_X_cluster.index = df_X_cluster_0.index

    return df_X_cluster


def coarse_grain_clone_over_cell_clusters(
    adata,
    selected_times=None,
    selected_fates=None,
    normalize=False,
    fate_normalize_source="X_clone",
    pseudocount=10 ** (-10),
    select_clones_with_fates: list = None,
    select_clones_without_fates: list = None,
    select_clones_mode: str = "or",
):
    """
    Compute the coarse-grained X_clone matrix over specified fates and
    time points, with the option to also normalize the resulting matrix
    first cluster-wise then clone-wise.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    selected_times: `list`, optional (default: None)
        Time points to select the cell states.
    selected_fates: `list`, optional (default: all)
        List of fate clusters to use. If set to be [], use all.
    normalize:
        To perform cluster-wise, then clone-wise normalization
    fate_normalize_source:
        Source for cluster-wise normalization: {'X_clone','state_info'}. 'X_clone': directly row-normalize coarse_X_clone; 'state_info': compute each cluster size directly, and then normalize coarse_X_clone. The latter method is useful if we have single-cell resolution for each fate.
    select_clones_with_fates: list = None,
        Select clones that labels fates from this list.
    select_clones_without_fates: list = None,
        Exclude clones that labels fates from this list.
    select_clones_mode: str = {'or','and'}
        Logic rule for selection.

    Returns:
    --------
    coarse_X_clone:
        The coarse-grained X_clone matrix

    mega_cluster_list:
         The updated cluster name list
    """

    time_info = np.array(adata.obs["time_info"])
    if selected_times is not None:
        if type(selected_times) is not list:
            selected_times = [selected_times]
    sp_idx = hf.selecting_cells_by_time_points(time_info, selected_times)
    X_clone_0 = adata[sp_idx].obsm["X_clone"]
    state_annote = adata[sp_idx].obs["state_info"]

    if fate_normalize_source not in ["X_clone", "state_info"]:
        raise ValueError("fate_normalize_source not in ['X_clone','state_info']")

    if np.sum(sp_idx) == 0:
        logg.error("No cells selected. Computation aborted!")
    else:
        (
            mega_cluster_list,
            valid_fate_list,
            fate_array_flat,
            sel_index_list,
        ) = hf.analyze_selected_fates(state_annote, selected_fates)
        if len(mega_cluster_list) == 0:
            logg.error("No cells selected. Computation aborted!")
        else:
            coarse_X_clone = np.zeros((len(mega_cluster_list), X_clone_0.shape[1]))
            for j, idx in enumerate(sel_index_list):
                coarse_X_clone[j, :] = X_clone_0[idx].sum(0)

        # map original fate name to mega_cluster name
        name_map = {
            list(set(state_annote.to_numpy()[x]))[0]: mega_cluster_list[j]
            for j, x in enumerate(sel_index_list)
        }

        # perform cluster-wise, then clone-wise normalization
        if normalize:
            # normalize cluster wise
            if fate_normalize_source == "state_info":
                logg.info("normalize by state_info (in case clonal data is sparse)")
                cell_type_N = np.array([np.sum(x) for x in sel_index_list])
                norm_X_cluster = coarse_X_clone / cell_type_N[:, np.newaxis]
            elif fate_normalize_source == "X_clone":
                logg.info("normalize by X_clone")
                norm_X_cluster = coarse_X_clone / (
                    coarse_X_clone.sum(1)[:, np.newaxis] + pseudocount
                )

            # normalize clone wise within each time point
            # first check if each cluster has only a single time point
            df_time_state = (
                pd.DataFrame(
                    {
                        "state_info": adata.obs["state_info"].to_list(),
                        "time_info": adata.obs["time_info"].to_list(),
                    }
                )
                .groupby(["state_info", "time_info"])
                .count()
                .reset_index()
            )
            df_time_state = df_time_state[
                df_time_state["state_info"].isin(fate_array_flat)
            ]
            df_time_state["mega_state"] = df_time_state["state_info"].map(name_map)

            each_cluster_has_unique_time = (
                df_time_state.groupby("state_info")["time_info"].count().to_numpy() == 1
            ).all()
            if each_cluster_has_unique_time:
                # each cluster has a single time point. normalize by time_info
                logg.info(
                    "each selected cluster has a unique time point. Normalize per time point"
                )
                X_list = []
                new_fate_list = []
                for t in sorted(df_time_state.time_info.unique()):
                    fates_t = df_time_state[df_time_state["time_info"] == t][
                        "mega_state"
                    ].unique()
                    sel_idx = np.in1d(mega_cluster_list, fates_t)

                    norm_X_cluster_clone_t = norm_X_cluster[sel_idx] / (
                        norm_X_cluster[sel_idx].sum(0)[np.newaxis, :] + pseudocount
                    )  # normalize by columns (i.e. clones)

                    X_list.append(norm_X_cluster_clone_t)
                    new_fate_list += list(mega_cluster_list[sel_idx])
                norm_X_cluster_clone = np.vstack(X_list)

                # re-order the matrix
                df_X_cluster = pd.DataFrame(norm_X_cluster_clone, index=new_fate_list)
                coarse_X_clone = df_X_cluster.loc[mega_cluster_list].to_numpy()

            else:
                logg.info(
                    "each cluster do not have a unique time point. Simply column-normalize the matrix"
                )
                norm_X_cluster_clone = norm_X_cluster / (
                    norm_X_cluster.sum(0)[np.newaxis, :] + pseudocount
                )  # normalize by columns (i.e. clones)

                coarse_X_clone = norm_X_cluster_clone

        if (select_clones_with_fates is not None) or (
            select_clones_without_fates is not None
        ):
            fate_names = np.array(mega_cluster_list)
            if select_clones_mode == "and":
                valid_clone_idx = np.ones(coarse_X_clone.shape[1]).astype(bool)
                if select_clones_with_fates is not None:
                    for x_name in select_clones_with_fates:
                        valid_clone_idx_tmp = (
                            coarse_X_clone[fate_names == name_map[x_name]].sum(0) > 0
                        )
                        valid_clone_idx = valid_clone_idx & valid_clone_idx_tmp

                if select_clones_without_fates is not None:
                    for x_name in select_clones_without_fates:
                        valid_clone_idx_tmp = (
                            coarse_X_clone[fate_names == name_map[x_name]].sum(0) > 0
                        )
                        valid_clone_idx = valid_clone_idx & ~valid_clone_idx_tmp
            elif select_clones_mode == "or":
                valid_clone_idx = np.zeros(coarse_X_clone.shape[1]).astype(bool)
                if select_clones_with_fates is not None:
                    for x_name in select_clones_with_fates:
                        valid_clone_idx_tmp = (
                            coarse_X_clone[fate_names == name_map[x_name]].sum(0) > 0
                        )
                        valid_clone_idx = valid_clone_idx | valid_clone_idx_tmp

                if select_clones_without_fates is not None:
                    for x_name in select_clones_without_fates:
                        valid_clone_idx_tmp = (
                            coarse_X_clone[fate_names == name_map[x_name]].sum(0) > 0
                        )
                        valid_clone_idx = valid_clone_idx | ~valid_clone_idx_tmp

            coarse_X_clone = coarse_X_clone[:, valid_clone_idx]

        return coarse_X_clone, mega_cluster_list


def get_normalized_coarse_X_clone(
    adata,
    selected_fates,
    fate_normalize_source="X_clone",
    pseudocount=10 ** (-10),
    select_clones_with_fates: list = None,
    select_clones_without_fates: list = None,
    select_clones_mode: str = "or",
):
    """
    We first normalize per cluster, then within each time point, normalize within the clone.
    In this case, the normalized coarse_X_clone matrix sums to 1 for each clone, thereby directly
    highlighting which cell type is more preferred by a clone. Note that the cluster-cluster correlation
    will be affected by both the cluster and clone normalization.

    Parameters
    ----------
    adata:
        adata object
    selected_fates:
        Selected cell clusters to perform coarse-graining
    fate_normalize_source:
        Source for cluster-wise normalization: {'X_clone','state_info'}. 'X_clone': directly row-normalize coarse_X_clone; 'state_info': compute each cluster size directly, and then normalize coarse_X_clone. The latter method is useful if we have single-cell resolution for each fate.
    """

    if fate_normalize_source not in ["X_clone", "state_info"]:
        raise ValueError(
            "fate_normalize_source must come from {'X_clone', 'state_info'"
        )

    coarse_X_clone, mega_cluster_list = coarse_grain_clone_over_cell_clusters(
        adata,
        selected_fates=selected_fates,
        normalize=True,
        fate_normalize_source=fate_normalize_source,
        pseudocount=pseudocount,
        select_clones_with_fates=select_clones_with_fates,
        select_clones_without_fates=select_clones_without_fates,
        select_clones_mode=select_clones_mode,
    )

    df_X_cluster = pd.DataFrame(
        coarse_X_clone,
        columns=[f"clone_{j}" for j in range(coarse_X_clone.shape[1])],
        index=mega_cluster_list,
    )

    return df_X_cluster


def clonal_trajectory(
    adata,
    selected_fates,
    fate_normalize_source="X_clone",
    select_clones_with_fates: list = None,
    select_clones_without_fates: list = None,
    select_clones_mode: str = "or",
    **kwargs,
):
    """
    Assign each cell a fate choice that belongs to its clone, and the
    clonal fate choice is determined by the coarse-grained and normalized
    barcode count matrix according to coarse_grain_clone_over_cell_clusters.
    The output is stored at adata.obs[f"clonal_traj_{sel_fates[j]}"]
    """

    df_X_cluster = coarse_grain_clone_over_cell_clusters(
        adata,
        selected_fates,
        normalize=True,
        fate_normalize_source=fate_normalize_source,
        select_clones_with_fates=select_clones_with_fates,
        select_clones_without_fates=select_clones_without_fates,
        select_clones_mode=select_clones_mode,
        **kwargs,
    )
    coarse_X_clone = df_X_cluster.to_numpy()
    X_clone = adata.obsm["X_clone"].A
    fate_map = np.zeros((adata.shape[0], coarse_X_clone.shape[0]))
    cell_id_list = []
    fate_list = []
    for j in np.arange(coarse_X_clone.shape[1]):
        sel_ids = np.nonzero(X_clone[:, j] > 0)[0]
        cell_id_list += list(sel_ids)
        for i in sel_ids:
            fate_list.append(
                list(coarse_X_clone[:, j])
            )  # fate choice of corresponding cells in the selected cell list

    fate_map[cell_id_list, :] = np.array(fate_list)
    sel_fates = list(df_X_cluster.index)
    for j, x in enumerate(fate_map.transpose()):
        adata.obs[f"clonal_traj_{sel_fates[j]}"] = x
        logg.info(f'Results saved at adata.obs[f"clonal_traj_{sel_fates[j]}"]')


def add_clone_id_for_each_cell(adata):
    clone_id = list(np.zeros(adata.shape[0], dtype=str))
    X_clone_old = adata.obsm["X_clone"].A
    for j in range(adata.shape[0]):
        clone_id[j] = ",".join(np.nonzero(X_clone_old[j, :])[0].astype(str))
    adata.obs["clone_id"] = clone_id
    logg.info("add information at obs['clone_id']")


def remove_multiclone_cell(adata, **kwargs):
    return filter_cells(adata, **kwargs)


def filter_cells(
    adata, clone_threshold=2, keep_discarded_clones=True, update_X_clone=True
):
    """
    cells with clone number >= clone_threshold will be removed (set to 0) from the X_clone.
    """
    barcode_N_per_cell = adata.obsm["X_clone"].sum(1).A.flatten()
    logg.info("Multiclone cell fraction:", (barcode_N_per_cell > 1).mean())
    logg.info(
        "Fraction of clones related to a multiclone cell:",
        (adata.obsm["X_clone"][barcode_N_per_cell > 1].sum(0) > 0).mean(),
    )
    X_clone_new = adata.obsm["X_clone"].A
    for x in np.nonzero(barcode_N_per_cell >= clone_threshold)[0]:
        for j in range(X_clone_new.shape[1]):
            X_clone_new[x, j] = 0

    absent_clone_before = np.nonzero(adata.obsm["X_clone"].A.sum(0) == 0)[0]
    lost_clones = np.nonzero(X_clone_new.sum(0) == 0)[0]
    removed_clones = set(lost_clones).difference(absent_clone_before)
    removed_clones = list(removed_clones)

    if keep_discarded_clones:
        logg.info("Add back dispeared clones due to cell removal")
        X_clone_new[:, removed_clones] = adata.obsm["X_clone"].A[:, removed_clones]

    adata.obsm["X_clone_old"] = adata.obsm["X_clone"]
    if update_X_clone:
        adata.obsm["X_clone"] = ssp.csr_matrix(X_clone_new)
        logg.info("Updated X_clone")
        logg.info(f"Check removed clones with adata.obsm['X_clone_old']")

    barcode_N_per_cell_new = X_clone_new.sum(1)
    df = pd.DataFrame(
        {
            "Clone_N_per_cell_old": barcode_N_per_cell,
            "Clone_N_per_cell_new": barcode_N_per_cell_new,
        }
    ).melt()
    fig, ax = plt.subplots()
    ax = sns.histplot(data=df, x="value", hue="variable")
    ax.set_xlabel("Clone number per cell")

    return removed_clones


def remove_multicell_clone(adata, **kwargs):
    return filter_clones(adata, **kwargs)


def filter_clones(adata, clone_size_threshold=2, filter_larger_clones=False):
    """
    if filter_larger_clones=True,
        clones with clone_size >= clone_size_threshold will be removed from the X_clone.
    else:
        filter smaller clones
    """
    clone_size = adata.obsm["X_clone"].sum(0).A.flatten()
    if filter_larger_clones:
        clone_idx = clone_size < clone_size_threshold
        logg.info(
            f"Removed clone with size >= {clone_size_threshold}; fraction {1-clone_idx.mean():.2f}"
        )
    else:
        clone_idx = clone_size >= clone_size_threshold
        logg.info(
            f"Removed clone with size < {clone_size_threshold}; fraction {1-clone_idx.mean():.2f}"
        )

    X_clone_new = adata.obsm["X_clone"][:, clone_idx]

    adata.obsm["X_clone_old"] = adata.obsm["X_clone"]
    if "clone_id" not in adata.uns:
        adata.uns["clone_id"] = np.arange(adata.obsm["X_clone"].shape[1])
    adata.uns["clone_id"] = np.array(adata.uns["clone_id"])[clone_idx]
    adata.obsm["X_clone"] = ssp.csr_matrix(X_clone_new)
    logg.info("Updated X_clone")
    logg.info(f"Check removed clones with adata.obsm['X_clone_old']")


def clone_statistics(adata, joint_variable="time_info", display_clone_stat=True):
    """
    Extract the number of clones and clonal cells for each time point
    """

    if "clone_id" not in adata.uns:
        adata.uns["clone_id"] = np.arange(adata.obsm["X_clone"].shape[1])
    adata.obs[joint_variable] = adata.obs[joint_variable].astype(str)

    df = (
        pd.DataFrame(adata.obsm["X_clone"].A)
        .reset_index()
        .rename(columns={"index": "cell_id"})
        .assign(time_info=list(adata.obs[joint_variable]))
        .set_index(["cell_id", "time_info"])
        .melt(ignore_index=False)
        .reset_index()
        .rename(columns={"variable": "clone_id", "value": "count"})
        .query("count>0")
    )

    df_cell = (
        df.groupby("time_info")
        .agg(cell_N=("cell_id", lambda x: len(set(x))))
        .assign(clonal_cell_fraction=lambda x: x["cell_N"] / x["cell_N"].sum())
    )

    df = df.groupby("clone_id").agg(
        cell_number=("cell_id", "count"),
        time_points=("time_info", lambda x: ",".join(list(set(x)))),
        time_point_N=("time_info", lambda x: len(set(x))),
    )

    df_clone = (
        df.groupby("time_points")
        .agg(clone_N=("time_points", "count"))
        .assign(clone_fraction=lambda x: x["clone_N"] / x["clone_N"].sum())
    )

    if display_clone_stat:
        logg.info(
            df_clone.reset_index()
            .rename(columns={"time_points": joint_variable})
            .set_index(joint_variable)
        )
        logg.info("-----------")
        logg.info(
            df_cell.reset_index()
            .rename(columns={"time_info": joint_variable})
            .set_index(joint_variable)
        )
    return df.reset_index().rename(
        columns={
            "time_point_N": f"{joint_variable}_N",
            "time_points": joint_variable,
            "cell_number": "clone_size",
        }
    )


def get_distance_within_each_clone(X_clone, norm_distance):
    # get distance vector for each clone
    t1_clone_size = (X_clone).sum(0)
    selected_clone_idx = np.nonzero(t1_clone_size >= 2)[0]
    distance_list = []
    for x in selected_clone_idx:
        ids_all = np.nonzero(X_clone[:, x] > 0)[0]
        distance_tmp = []
        for i in range(len(ids_all)):
            distance_tmp.append(
                [
                    norm_distance[ids_all[i], ids_all[j]]
                    for j in range(len(ids_all))
                    if j != i
                ]
            )
        distance_list.append(np.array(distance_tmp).flatten().max())  # min.mean
    return distance_list, selected_clone_idx


def compute_sister_cell_distance(
    adata,
    selected_time=None,
    method="2D",
    precomputed_distance_matrix=None,
    key="X_emb",
    neighbor_number=10,
    title=None,
    color_random="#a6bddb",
    color_data="#fdbb84",
    max_N_simutation=1000,
    alternative="less",
):
    """
    Parameters
    ----------
        selected_time:
            selected time point
        method:
            {'2D','high'} using 2D embedding or high-dimension state measurement X_pca.
        key:
            If method='2D', decide which 2-D embedding to use. Could be 'X_umap', or 'X_emb' etc.
        neighbor_number:
            If method='high', decide the neighbor_number to construct the KNN graph.
        max_N_simutation:
            Max number to perform simulation to generate a null distribution. A higher number may give more stronger
            statistical statement.
        alternative:
            'less', 'greater','two-sided'

    Returns
    -------
        df_distance:
            A dataframe for cell-cell distance
    """

    from scipy.spatial.distance import pdist, squareform

    if selected_time is None:
        adata_t1 = adata
    else:
        adata_t1 = adata[adata.obs["time_info"] == selected_time]

    if precomputed_distance_matrix is not None:
        logg.info("use precomputed distance matrix")
        norm_distance = precomputed_distance_matrix
    else:
        if method == "2D":
            logg.info("Use 2-dimension embedding")
            X = adata_t1.obsm[key]
            norm_distance = squareform(pdist(X))

        elif method == "high":
            logg.info("Use high-dimension")
            norm_distance = hf.compute_shortest_path_distance(
                adata_t1,
                normalize=False,
                num_neighbors_target=neighbor_number,
            )
        else:
            raise ValueError("method should be among {'2D' or 'high'}")

    X_clone = (adata_t1.obsm["X_clone"].A.copy() > 0).astype(int)
    logg.info(np.sum(X_clone.sum(0) >= 2), " clones with >=2 cells selected")

    # observed distances
    distance_list, selected_clone_idx = get_distance_within_each_clone(
        X_clone, norm_distance
    )

    # randomized distances
    random_dis = []
    random_dis_stat = []
    for _ in tqdm(range(max_N_simutation)):
        np.random.shuffle(X_clone)
        temp, __ = get_distance_within_each_clone(X_clone, norm_distance)
        random_dis += temp
        random_dis_stat.append(
            [np.mean(temp), np.min(temp), np.median(temp), np.max(temp)]
        )
    random_dis_stat = np.array(random_dis_stat)

    df_obs = pd.DataFrame(
        {
            "clone_id": selected_clone_idx,
            "distance": distance_list,
        }
    )
    df_obs["source"] = "Observed"

    df_rand = pd.DataFrame(
        {
            "clone_id": np.arange(len(random_dis)).astype(str),
            "distance": random_dis,
        }
    )
    df_rand["clone_id"] = "rand_" + df_rand["clone_id"]
    df_rand["source"] = "Random"

    df_distance = pd.concat([df_obs, df_rand], ignore_index=True)
    df_distance["source"] = pd.Categorical(df_distance["source"]).set_categories(
        ["Random", "Observed"], ordered=True
    )

    bins = np.linspace(df_distance["distance"].min(), df_distance["distance"].max(), 50)
    ax = sns.histplot(
        data=df_distance[df_distance["source"] == "Random"],
        label="Random",
        bins=bins,
        stat="probability",
        color=color_random,
    )
    ax = sns.histplot(
        data=df_distance[df_distance["source"] == "Observed"],
        x="distance",
        label="Observed",
        bins=bins,
        stat="probability",
        color=color_data,
        alpha=0.5,
    )
    ax.legend()
    ax.set_xlabel("Sister-cell distance")

    stat_score, pvalue = stats.ranksums(
        df_obs["distance"].to_numpy(), random_dis, alternative=alternative
    )

    if title is None:
        ax.set_title(f"t={selected_time}, Pvalue: {pvalue:.2f}")
    else:
        ax.set_title(title)
    plt.tight_layout()
    if "data_des" in adata.uns.keys():
        data_des = adata.uns["data_des"][-1]
    else:
        data_des = ""
    plt.savefig(f"{settings.figure_path}/transcriptome_memory{data_des}.pdf")

    return df_distance, pvalue


def generate_adata_from_X_clone(X_clone, state_info=None, time_info=None):
    """
    Convert X_clone matrix to adata, and also add it to adata.obsm['X_clone'].
    You can run cospar on it directly.
    """
    import scanpy as sc

    adata_orig = sc.AnnData(X_clone)
    adata_orig.obsm["X_clone"] = X_clone
    adata_orig.uns["data_des"] = ["hi"]
    if state_info is None:
        adata_orig.obs["state_info"] = pd.Categorical(
            np.arange(X_clone.shape[0]).astype(str)
        )
    else:
        adata_orig.obs["state_info"] = pd.Categorical(state_info)

    if time_info is None:
        adata_orig.obs["time_info"] = ["0"] * X_clone.shape[0]
    else:
        adata_orig.obs["time_info"] = pd.Categorical(time_info)
    return adata_orig


def conditional_heatmap(
    coarse_X_clone,
    fate_names: list,
    time_info=None,
    included_fates: list = None,
    excluded_fates: list = None,
    binarize=False,
    normalize=False,
    mode="and",
    fig_height=1.3 * plt.rcParams["figure.figsize"][0],
    fig_width=plt.rcParams["figure.figsize"][0],
    **kwargs,
):
    """
    Plot a heatmap by conditionally including or removing a set of fates

    log_transform: True or False
    """
    fate_names = np.array(fate_names)

    if mode == "and":
        valid_clone_idx = np.ones(coarse_X_clone.shape[1]).astype(bool)
        if included_fates is not None:
            for x_name in included_fates:
                valid_clone_idx_tmp = coarse_X_clone[fate_names == x_name].sum(0) > 0
                valid_clone_idx = valid_clone_idx & valid_clone_idx_tmp

        if excluded_fates is not None:
            for x_name in excluded_fates:
                valid_clone_idx_tmp = coarse_X_clone[fate_names == x_name].sum(0) > 0
                valid_clone_idx = valid_clone_idx & ~valid_clone_idx_tmp
    elif mode == "or":
        valid_clone_idx = np.zeros(coarse_X_clone.shape[1]).astype(bool)
        if included_fates is not None:
            for x_name in included_fates:
                valid_clone_idx_tmp = coarse_X_clone[fate_names == x_name].sum(0) > 0
                valid_clone_idx = valid_clone_idx | valid_clone_idx_tmp

        if excluded_fates is not None:
            for x_name in excluded_fates:
                valid_clone_idx_tmp = coarse_X_clone[fate_names == x_name].sum(0) > 0
                valid_clone_idx = valid_clone_idx | ~valid_clone_idx_tmp

    adata = generate_adata_from_X_clone(
        ssp.csr_matrix(coarse_X_clone[:, valid_clone_idx]),
        state_info=fate_names,
        time_info=time_info,
    )

    pl.barcode_heatmap(
        adata,
        normalize=normalize,
        binarize=binarize,
        selected_fates=fate_names,
        order_map_x=False,
        order_map_y=True,
        fig_height=fig_height,
        fig_width=fig_width,
        **kwargs,
    )
    X_count = adata.uns["barcode_heatmap"]["coarse_X_clone"]

    return X_count
